package heasarc.w3browse;

import java.net.*;
import java.util.Vector;

import java.io.*;
import heasarc.w3browse.plot.*;
import heasarc.util.DoubleVector;


/** This class queries the HEASARC database.  Results are
  * returned as an array of columns.  The query is run in a
  * separate thread.  The object tries to parse header information
  * from the query quickly so that column information is available
  * even before all data has been read in.  However no actual query
  * results are returned until the entire query has been parsed.
  */
public class HeasarcQuery implements Runnable {
    
    /** A marker for the beginning of the data */
    private static String dataStart="BatchStart";
    
    /** A marker for the end of the data */
    private static String dataEnd="BatchEnd";
    
    /** A reader connected to the query URL */
    private BufferedReader urlData;
    
    /** The thread processing the query */
    private Thread	t;
    
    /** Is the query data ready? */
    private boolean	ready   = false;
    
    /** Has there been a problem reading or parsing the data? */
    private boolean	badData = false;
    
    
    /** Has the header information been read? */
    private boolean	headerReady = false;
    
    /** Vectors used in accumulating data. */
    private Object[] 	colvec;
    
    /** Array of columns of data */
    private Column[]	dataCols;
    
    /** Names of columns */
    private String[]	colNames;
    
    /** Types of columns */
    private String[]	colTypes;
    
    /** Are the columns numeric? */
    private boolean[]	isColNum;
    
    /** The constructor starts the query.
      *
      */
    public HeasarcQuery (String fullUrl, String data) throws IOException {
	
	URL    queryUrl;
	URLConnection uc;
        //System.out.println("HeasarcQuery::fullUrl=" + fullUrl);
	
	try {
	    queryUrl = new URL(fullUrl);
	    uc = queryUrl.openConnection();
	    uc.setDoOutput(true);
	    uc.setDoInput(true);
	    uc.setAllowUserInteraction(false);
	    DataOutputStream dos = new DataOutputStream(uc.getOutputStream());
	    dos.writeBytes(data);
	    dos.close();
	} catch (Exception e) {
	    throw new java.io.IOException("Error opening URL:"+fullUrl);
	}
	
	InputStream is = uc.getInputStream();
	urlData = new BufferedReader(
		      new InputStreamReader(is) );
	
	// Now Start a thread to read the data.
	this.t = new Thread(this);
	t.start();
    }
    
    
    /** Read and parse the query data.  This is run
      * in a separate thread to allow a program to
      * do other things while a query is processed.
      */
    public void run () {
	
	try {
	    
	    String line;
	    
	    while (true) {
		line = urlData.readLine();
		if (line == null || line.equals(dataStart)) {
		    break;
		}
	    }
	    
	    
	    
	    getNames(urlData.readLine());
	    getTypes(urlData.readLine());
	    
	    int nrow = 0;
	    while (true) {
		 
		line = urlData.readLine();
		if (line == null || line.equals(dataEnd)) {
		    break;
		}
		getData(line);
		nrow += 1;
	    }
	    
	} catch (IOException e) {
	    badData = true;
	}
	
	if (!badData) {
	    finishData();
	}
    }
    
    /** Get the names */
    private void getNames(String line) {
	
	
	Vector vnames = new Vector(32);
	int    pos   = 1;  // Skip initial |
	
	if (line == null) {
	    badData = true;
	    return;
	}
	
	while (pos < line.length()) {
	    int end = line.indexOf('|', pos);
	    if (end < 0) {
		break;
	    }
	    vnames.addElement(line.substring(pos,end));
	    pos = end+1;
	}
	
	colNames = new String[vnames.size()];
	vnames.copyInto((Object[])colNames);
	
    }
    
    
    /** Get the Types */
    private void getTypes(String line) {

  	if (line == null) {
	    badData = true;
	    return;
	}
	
	colTypes = new String[colNames.length];
	
	int pos = 1;    // Skip initial |
	
	for (int i=0; i < colTypes.length; i += 1) {
	    
	    int end = line.indexOf('|', pos);
	    if (end < 0) {
		break;
	    }
	    colTypes[i] = line.substring(pos,end).toUpperCase();
	    pos = end+1;
	}
	
	colvec    = new Object[colNames.length];
	isColNum  = new boolean[colNames.length];
	
	for (int i=0; i<colNames.length; i += 1) {
	    if (colTypes[i].equals("1") || colTypes[i].equals("-1") ) {
		colvec[i]     = new Vector(256);
		isColNum[i]   = false;
	    } else {
		colvec[i]     = new DoubleVector();
		isColNum[i]   = true;
	    }
	}
	
	headerReady = true;
    }
    
    /** Get the data */
    private void getData(String line) {
	
	int pos = 1;
	for (int i=0; i < colNames.length; i += 1) {
	    
	    int end = line.indexOf('|', pos);
	    if (end < 0) {
		break;
	    }
	    
	    if (!isColNum[i]) {
	        ((Vector)(colvec[i])).addElement(line.substring(pos,end));
	    } else {
		DoubleVector v = (DoubleVector) colvec[i];
	        if (end == pos) {
		    v.add(Double.NaN);
		} else {
		    try {
		        v.add(new Double(line.substring(pos,end)).doubleValue());
		    } catch(Exception e) {
			v.add(Double.NaN);
		    }
		}
	    }
	    pos = end+1;
	}
    }
    
    /** Finish up information on each column.
     */
    private void finishData() {
	
	dataCols = new Column[colNames.length];
	
	for (int i=0; i<colNames.length; i += 1) {
	    Object data;
	    
	    if (isColNum[i]) {
		data = ((DoubleVector)(colvec[i])).toArray();
	    } else {
		data = new String[((Vector)colvec[i]).size()];
	        ((Vector)(colvec[i])).copyInto((Object[])data);
	    }
	    dataCols[i] = new Column(colNames[i], colTypes[i], data,
				     isColNum[i]);
            //System.out.println( "HQ:finishdata: new Column for " + 
		//colNames[i] +", " + colTypes[i]);
	}
	ready = true;
    }
		

    /** Get the number of columns.
      * This routine will return as soon as
      * the header is parsed.
      */
    public int getNcol() throws BadDataException {
	
	waitForHeaderReady();
	
        if (badData) {
          throw new BadDataException("HeasarcQuery object contains bad data");
        }
	
	return colNames.length;
    }


    /** Get a column.  This routine will return after
      * the entire query has been parsed.
      * @param col The index of the desired column
      */
    public Column getColumn(int col) throws BadDataException {
	
	waitForReady();
	if (col < 0 || col >= colNames.length) {
	    return null;
	} else {
	    return dataCols[col];
	}
    }

    /** Get a column.  This routine will return after
      * the entire query has been parsed.
      * @param col The name of the desired column
      */
    public Column getColumn(String colname) throws BadDataException {
	
	waitForReady();
        for(int i=0;i<colNames.length;i++) {
          if(colNames[i].equals(colname)) {
             return (dataCols[i]);
          }
        }
        throw new BadDataException("HeasarcQuery object does not contain " + colname);
    }


    /** Wait until parsing is complete */
    private void waitForReady() throws BadDataException {
	
	while (!ready && t != null && t.isAlive()) {
	    try {
		Thread.sleep(100);  // Sleep for 0.1 seconds.
	    } catch(InterruptedException e) {
	    }
	}
	if (badData) {
	    throw new BadDataException("HeasarcQuery request did not complete");
	}
	  
	return;
    }
    
    /** Wait until header parsing is complete */
    private void waitForHeaderReady() {
	
	while (!headerReady && t != null && t.isAlive()) {
	    
	    try {
		Thread.sleep(100);  // Sleep for 0.1 seconds.
	    } catch(InterruptedException e) {
	    }
	    
	}
	return;
    }
    
    /** Get the column names.  This routine will
      * return valid data as soon as header parsing is complete.
      * @param wait Wait for the results?
      */
    public String[] getNames(boolean wait) throws BadDataException {
	
	if (headerReady) {
	    
            if (badData) {
               throw new BadDataException("HeasarcQuery object contains bad data");
            }
	    
	    return colNames;
	    
	} else if (wait) {
	    waitForHeaderReady();
            if (badData) {
               throw new BadDataException("HeasarcQuery object contains bad data");
            }
	    return colNames;
	    
	} else {
	    return null;
	}
    }
    
    /** Get the column types.  This routine can return valid results
      * as soon as header parsing is complete.
      * @param wait Wait for results?
      */
    public String[] getTypes(boolean wait) throws BadDataException {
	
	if (headerReady) {
            if (badData) {
               throw new BadDataException("HeasarcQuery object contains bad data");
            }
	    return colTypes;
	} else if (wait) {
	    waitForHeaderReady();
            if (badData) {
               throw new BadDataException("HeasarcQuery object contains bad data");
            }
	    return colTypes;
	} else {
	    return null;
	}
    }
    
    /** Get flags indicating numeric columns.  This routine
      * can return valid data as soon as header parsing is complete.
      * @param wait Wait for results?
      */
    public boolean[] getIsNumeric(boolean wait) throws BadDataException {
	
	if (headerReady) {
	    
            if (badData) {
               throw new BadDataException("HeasarcQuery object contains bad data");
            }
	    return isColNum;
	    
	} else if (wait) {
	    
	    waitForHeaderReady();
            if (badData) {
               throw new BadDataException("HeasarcQuery object contains bad data");
            }
	    return isColNum;
	    
	} else {
	    
	    return null;
	}
    }
    
    /** Did the query fail?
      */
    public boolean isFail() {
	try {
	    waitForReady();
	} catch (Exception e) {
	}
	return badData;
    }
    
    public int getNrow() throws BadDataException {
	waitForReady();
	if (colNames.length == 0) {
	    throw new BadDataException("No columns in query");
	}
	return dataCols[0].getNrows();
    }
    
    /** Get all columns */
    public Column[] getColumns() throws BadDataException {
	waitForReady();


	Column[] data = new Column[colNames.length];
	for (int i=0; i<colNames.length; i += 1) {
	    data[i] = getColumn(i);
	}
	return data;
    }
    
    /** Get numeric columns */
    public Column[] getNumericColumns() throws BadDataException {
	waitForReady();

	int count=0;
	for (int i=0; i<colNames.length; i += 1) {
	    if (isColNum[i]) {
		count += 1;
	    }
	}
	
	Column[] data = new Column[count];
	
	count = 0;
	for (int i=0; i<colNames.length; i += 1) {
	    if (isColNum[i]) {
		data[count] = getColumn(i);
		count += 1;
	    }
	}
	return data;
    }
    
    /** Get numeric column names */
    public String[] getNumericNames() throws BadDataException {

        waitForHeaderReady();
        if (badData) {
           throw new BadDataException("HeasarcQuery object contains bad data");
        }

        int count=0;
        for (int i=0; i<colNames.length; i += 1) {
            if (isColNum[i]) {
                count += 1;
            }
        }

        String[] data = new String[count];

        count = 0;
        for (int i=0; i<colNames.length; i += 1) {
            if (isColNum[i]) {
                data[count] = colNames[i];
                count += 1;
            }
        }
        return data;
    }

}

/** BadDataException */
class BadDataException extends Exception {
  /** constructor for BadDataException */
  public BadDataException(String message) {
     super(message);
  }
  /** constructor for BadDataException */
  public BadDataException() {
     super();
  }

}

