# $Id: CSIO.pm,v 1.4 2002/01/11 22:04:52 cwh Exp $
# $Log: CSIO.pm,v $
# Revision 1.4  2002/01/11 22:04:52  cwh
# Added an error check.
#
# Revision 1.3  2001/11/07 23:34:16  cwh
# Moved everything except DBD/ into lib/Browse/, so as not to pollute the module
# namespace with our Sort, Search, CGI, etc.  This required putting Browse:: in
# front of every module call and package declaration.  Also any mod::new() had
# to be changed to mod->new().
#
# Revision 1.2  2000/07/26 18:15:31  saima
# Minor change to error message.
#
# Revision 1.1  1999/05/27 16:57:12  tam
# Initial CVS entry
#

package Browse::CSIO;


# This package emulates Song Yom's CSIO C package.  It comprises
# two sections.  The first part consists of routines which
# provide socket connections and the second is a set of I/O utilities
# constructed to ensure synchronization of the client and server.
#
# Server usage:
#   Call server(port, routine)
#   where the port number and a routine to be executed for each
#   connection should be supplied.
#
# Client usage:
#   The client(host, port) function should be called.  It will
#   return a pointer to the socket used for I/O.  
#
# Last Modified:
#  1999-02-05  TAM
#              Fixed bug in client routine.  It wasn't dupping
#              the Socket to the CS_INPUT and CS_OUTPUT descriptors.
#  1998-03-05  TAM
#              Fixed bug in inp() reading long messages
#  1998-02-06  TAM
#              Eliminated references to REAPER -- these can interrupt
#              I/O.  Reaped processes explicitly in accept loop.
#  1998-02-05  TAM 
#              Modified REAPER to call waitpid after wait
#              to check for additional processes that may have died.
#              Made output to Log file unbuffered.
#              Modified format of log file entries.
#              Made log

use strict;
use Socket;
use Fcntl;


sub spawn;

# Log a message.
sub logmsg {
    $| = 1;
    print substr(scalar localtime,4), ":$$-", @_ , "\n";
}

# Statistics stuff.
my ($numCall, $totalTime, $longest);

my %startTimes;
sub getStats {

    return ($numCall, $totalTime, $longest);
}


# Reaps dead child processes.
# This routine is not currently used.  Children
# are reaped explicitly rather than by
# catching the SIGCHLD signal.

sub REAPER  {

    $SIG{CHLD} = \&REAPER;
    my $waitedpid = wait;
    
    if (exists($startTimes{$waitedpid})) {
        my $duration = time - $startTimes{$waitedpid};
        if ($duration > $longest) {
            $longest = $duration;
        }
        $totalTime += $duration;
	$numCall += 1;
        delete($startTimes{$waitedpid});
        logmsg "Exit $waitedpid". ($? ? " with exit $?" : "");
    }
    # Wait for other processes.
    my $id;
    while (($id=waitpid(-1,1)) > 0) {
    }
    
    
}

# Set up a server socket and wait for connections.

sub server {

    # Only process children at the end of the Accept loop
    $SIG{CHLD} = 'IGNORE';

    # Set up server side of client server.
    
    my ($port, $routine) = @_;

    my $proto = getprotobyname('tcp');
    


    # Taken from Perl book page 350.
    socket(Server, PF_INET, SOCK_STREAM, $proto)                or die "socket: $!";
    setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))  or die "setsockopt $!";
    bind(Server, sockaddr_in($port, INADDR_ANY))                or die "bind: $!";
    listen(Server, SOMAXCONN)                                   or die "listen: $!";
    logmsg "Server started on port: $port";

    my $waitedpid = 0;
    my $paddr;

   
#    $SIG{CHLD} = \&REAPER;
    
    # Loop waiting for connections.

    for (; $paddr = accept(Client, Server);   close (Client) ) {
    
        # Ignore interrupts until we have all the system stuff set up.
        $SIG{CHLD} = 'IGNORE';
    
	# Now clean up any dead children before we start.
	my $id;
        while (($id=waitpid(-1,1)) > 0) {
            logmsg(sprintf("%d %x", $id, $?));
	}
        my $stat = fcntl(Client, F_SETFL, O_SYNC);
        my ($port, $iaddr) = sockaddr_in($paddr);
        my $name = gethostbyaddr($iaddr, AF_INET);
        logmsg "Connect from $name [", inet_ntoa($iaddr), "] / $port";
        spawn($routine);
	
    }
}

# Start a child to handle a client connection to the server.

sub spawn {

    my ($routine) = @_;
    
    my $pid;
    if (!defined($pid = fork())) {
        logmsg "Cannot fork";
        return;
    }
    if ($pid) {
        logmsg "Begat: $pid";
	$startTimes{$pid} = time;
	
	# Restore interrupts.
#	$SIG{CHLD} = \&REAPER;
	return;
    }
    
    open(CS_INPUT, "<&Client") or die "Can't dup to stdin";
    open(CS_OUTPUT, ">&Client") or die "Can't dup to stdout";
    select CS_INPUT;
    $| = 1;
    select(CS_OUTPUT);
    $| = 1;
    select(STDERR);
    $| = 1;
    select(STDOUT);
    &$routine;
    exit;
}

# Connect to a server.  This code has not
# been extensively tested.sub client {
sub client {

    # Establish a client connection to a port.  This
    # routine returns with the bi-directional socket.
    
    # Code adapted from Perl book p349.
    
    my ($remote, $port) = @_;
    
    my ($iaddr, $paddr, $proto, $line);
    
    if ($port =~ /\D/) {
        $port = getservbyname($port, 'tcp');
    }
    die "Browse error: CSIO::client No port" unless $port;
    
    $iaddr = inet_aton($remote);
    $paddr = sockaddr_in($port, $iaddr);
    $proto = getprotobyname('tcp');
    
    socket(SOCK, PF_INET, SOCK_STREAM, $proto) or die "socket $1";
   
    connect(SOCK, $paddr) or die "Can't connect to $remote";
   
    open(CS_INPUT, "<&SOCK") or die "Can't dup to stdin";
    open(CS_OUTPUT, ">&SOCK") or die "Can't dup to stdout";
    return;
}

# This implements Song Yom's protocol for client-server
# connections described below.
#
# Each communication passed between the client and server has a fixed
# format comprising a 3 character header which indicates whether
# this is a message or data, followed by a 12 character field giving
# the length of the message/data in ASCII.  This field is right justified
# and is null filled.  The message/data starts in the sixteenth byte.
# For data which exceeds the maximum buffer size, the data is
# broken into sections of the maximum buffer size and all but the
# first are transmitted without a header or length section.  However,
# splitting up is supported only on reads, not on writes.
#
# The protocol is as follows:
#
#    Initially the server sends the client an SSTAT message
#    The client should respond with an SSIGN message
#    Server Output:
#         Send data
#         Listen for an message acknowledgment with the SSIGN message
#    Server Input:
#         Send the SSIGN message
#         Read the data
#    Client Output:
#         Read from server.
#         If this was data
#               Send acknowledgement
#               Read again (should be a message)
#         Send data
#    Client input:
#         Read message
#         Send SSIGN
#
# Note that the client may skip data messages to synchronize with server.
#


$Browse::CSIO::maxBuffer = 4096;
$Browse::CSIO::lenLength = 12;
$Browse::CSIO::msgLength = 3;

$Browse::CSIO::dsig = '<d>';		# Data
$Browse::CSIO::msig = '<m>';		# Message

$Browse::CSIO::DFLAG = 1;
$Browse::CSIO::MFLAG = 2;

$Browse::CSIO::errorStatus = -2;
$Browse::CSIO::fatalStatus = -3;
$Browse::CSIO::okStatus = 1;
$Browse::CSIO::inpLength = $Browse::CSIO::maxBuffer + $Browse::CSIO::lenLength + $Browse::CSIO::msgLength;

$Browse::CSIO::SSTAT = "1\0";    # Startup message.
$Browse::CSIO::SSIGN = "2\0";    # Acknowledgment/I'm ready message


# Send output.
sub out {

    my ($buffer, $isData) = @_;
    
    my $prefix;
    
    if ($isData != 0) {
        $prefix = $Browse::CSIO::dsig;
    } else {
        $prefix = $Browse::CSIO::msig;
    }        
    
    # Note the silent truncation of long buffers.  This
    # doesn't correspond to the inp routine which handles data
    # longer than the buffer size.
    
    my $dataLength = length($buffer);
    if ($dataLength > $Browse::CSIO::maxBuffer) {
	$dataLength = $Browse::CSIO::maxBuffer;
    }

    my $lengthField = sprintf("%d", $dataLength) . ("\0" x $Browse::CSIO::lenLength);
    $lengthField = substr($lengthField, 0, $Browse::CSIO::lenLength);
    
    my $outlen =  $Browse::CSIO::msgLength+$Browse::CSIO::lenLength+$dataLength;
    my $status = syswrite(CS_OUTPUT, 
                   $prefix.$lengthField.substr($buffer, 0, $dataLength), 
		   $outlen);
		   

    if ($status != $outlen) {
        return $Browse::CSIO::errorStatus;
    } else {
        return $Browse::CSIO::okStatus;
    }
}	

# Get input.
sub inp {

    my $buffer;
    my $length;
    my $flag;
    
    
    $length = sysread(CS_INPUT, $buffer, $Browse::CSIO::inpLength);
    if ($length > 0) {
    
        # Is this data or message?
	
        my $prefix = substr($buffer, 0, $Browse::CSIO::msgLength);
	if ($prefix eq $Browse::CSIO::msig) {
	    $flag = $Browse::CSIO::MFLAG;
	} elsif ($prefix eq $Browse::CSIO::dsig) {
	    $flag = $Browse::CSIO::DFLAG;
	}
	
	# Get the length of the message.
        my $need = substr($buffer, $Browse::CSIO::msgLength, $Browse::CSIO::lenLength);

	
	# Get the beginning of the message.
	
	my $buffer = substr($buffer, $Browse::CSIO::msgLength+$Browse::CSIO::lenLength);
	
	
	# If we need any further data read it in.
	$need -= length($buffer);
	while ( $need > 0 ) {
	    my $newbuf;
	    my $newlen = sysread(CS_INPUT, $newbuf, $need);
	    if ($newlen <= 0) {
	        return $Browse::CSIO::errorStatus;
	    }
	    
	    $need -= $newlen;
	    $buffer .= $newbuf;
	}
	
	if ($need != 0 || !$flag) {
	    return $Browse::CSIO::errorStatus;
	}
	
	return ($Browse::CSIO::okStatus, $buffer, $flag);
	
    } else {
    
        return $Browse::CSIO::errorStatus;
    }
}

# Handle the output protocol.

sub output {

    my ($buffer, $isServer) = @_;
    my $fbuf;
    
    if (!$isServer) {
        # Read the status message if this is a client.
    
        my ($status, $input, $inflag) = inp();
	if ($status != $Browse::CSIO::okStatus) {
	    return $Browse::CSIO::errorStatus;
	}
	
	if (substr($input, 0, $Browse::CSIO::msgLength) eq $Browse::CSIO::dsig) {
	    # Server is sending data.  Read in this message, ack, and
	    # try with the next.
	    if (out($Browse::CSIO::SSIGN, 0)  != $Browse::CSIO::okStatus) {
	        return $Browse::CSIO::errorStatus;
	    }
	    ($status, $input, $inflag) = inp();
	    if ($status != $Browse::CSIO::okStatus) {
	        return $Browse::CSIO::errorStatus;
	    }
	}
    }

    # Send the data
    if (out($buffer, 1)  != $Browse::CSIO::okStatus) {
        return $Browse::CSIO::errorStatus;
    }
    
    if ($isServer) {
    
        # Read the acknowledgement.
	
        my ($status, $buffer, $flag) = inp();
	
	if ($status != $Browse::CSIO::okStatus || $buffer ne $Browse::CSIO::SSIGN  ||
	    $flag != $Browse::CSIO::MFLAG) {
	    return $Browse::CSIO::errorStatus;
	}
    }
    
    return $Browse::CSIO::okStatus;
    
}
	      
# Handle the input protocol.    
sub input {

    my ($isServer) = @_;
    
    if ($isServer) {
        
        # Send sync message.
        if (out($Browse::CSIO::SSIGN, 0) != $Browse::CSIO::okStatus) {
	    return $Browse::CSIO::errorStatus;
	}
	
    }
    
    my ($status, $input, $flag) = inp();
    
    if (!$isServer) {
        if (out($Browse::CSIO::SSIGN, 0) != $Browse::CSIO::okStatus) {
	    return $Browse::CSIO::errorStatus;
	}
    }
    if ($flag == 2) {
        undef($input);
    }
    
    
    return ($Browse::CSIO::okStatus, $input);
}
    
sub sendingData {


    # Send out a signal indicating we are about to send a data stream.
    
    if (out($Browse::CSIO::SSTAT, 0) == $Browse::CSIO::okStatus) {
    
        my ($status, $data, $flag) = inp();
	
	if ($status != $Browse::CSIO::okStatus || $data ne $Browse::CSIO::SSIGN  ||
	    $flag != $Browse::CSIO::MFLAG) {
	    return $Browse::CSIO::errorStatus;
	}
	
    }
    return $Browse::CSIO::okStatus;
}

sub lookingForData {

    # Look for indication that we are going to get data.

    
    my ($status, $input, $flag) = inp();
    
    if ($status != $Browse::CSIO::okStatus || $input != $Browse::CSIO::SSTAT  ||
        $flag != $Browse::CSIO::MFLAG) {
	return $Browse::CSIO::errorStatus;
    }
    
    return out($Browse::CSIO::SSIGN, 0);
}

1;
