# $Id$

use strict;

use Browse::Results;
use Browse::ValidateInput;

# Standard display object.
# This class, the default Display object,
# returns a page with results formatted in HTML tables.

package Browse::CrossDisplay;

use vars qw( @ISA );
@ISA = qw( Browse::Display );

# if row given, saves a copy of the last row seen
# else returns the last row seen
# tdesignation is either 'primary' or 'secondary' and is required.

sub lastRow {
  my($self,$tdesignation,$row) = @_;
  return unless $tdesignation =~ /primary|secondary/i;
  if ($row) {
     $$self{lastrow_$tdesignation} = $row; 
     return;
  } else {
     return $$self{lastrow_$tdesignation};
  }
}


# Print the information at the beginning of a tables results.

sub tableHeader {
    my ($self) = @_;

    my ($meta) = $$self{"meta"};
    Browse::Results::DisplayLinks($meta);
    
    my $table_info = $$meta{"tableinfo"};
    my $table      = $$meta{"table"};
    
    $$self{"meta"} = $meta;
    $$self{"rowcount"} = 0;
    
    
    my ($sql) = $$meta{"sql"};
    
    if (defined($$table_info{"default_search_radius"})) {

	my $raf  = $$table_info{"right_ascension"};
	my $decf = $$table_info{"declination"};

	if (substr($raf, 0, 1) eq "@") {
	    $raf = substr($raf, 1);
	}
	if (substr($decf, 0, 1) eq "@") {
	    $decf = substr($decf, 1);
	}
	    
    }
	
	
    my $radius;
    if ($$meta{"positions"}) {
	my $pos = $$meta{"positions"};
	my $pos1 = $$pos[0];
	my ($ra, $dec, $rad) = split('\|', $pos1);
	$radius = $rad;
    }
    
    if ($$meta{"resource_name"} && $$meta{"resource_name"} ne $$self{"last_resource"}) {
        my $url = $$meta{"query_url"};
	$url =~ s/asu-xml/VizieR/;
        print "<hr><img alt=\"VizieR logo\" src='$::vizgif' height=20 width=20>  <a target=helpwin href='$url'>$$meta{resource_title} ($$meta{resource_name})</a>\n";
	$$self{"last_resource"} = $$meta{"resource_name"};
    }

    Browse::Results::PrintTableName($table, $table_info, $radius);

}

# Extract a specified set of fields from a table
# Finish off any information for a table.

sub tableFooter {
    my ($self,$msg) = @_;
  
    $self->closeTable($msg);
    
}

# Process a metadata input.  This implies that
# information from a new table will be forthcoming.


sub processMeta {
    my ($self, $meta) = @_;
    
    $self->closeTable();
    
    # Just save the metadata for later use.
    $$self{"meta"} = $meta;
    $$self{"rowcount"} = 0;
    
    # Make sure we don't hold onto an invalid transform.
    $$self{"transform"} = undef;
    
}

# Process a row of data.

sub processData {
    
    my ($self, $data) = @_;
    my $row = $$self{"rowcount"};

    if (!$$self{"transform"}) {
        $$self{"transform"} = $self->setTransform($data);
    }
    
    #if ($row % 25 == 0)  {
    if ($row == 0)  {
    
        if ($$self{"intable"}) {
            print "</tbody>\n";
	    print "</table>\n";
	}

        my $names = $$self{names};
        my $meta      = $$self{meta};
        my $coordsys  = $$self{coordsys};

        if ($coordsys !~ /equatorial/i) {
            my $tableinfo = $$meta{tableinfo};

            my $ra = $$tableinfo{"right_ascension"};
            $ra =~ s/^\@//;
            my $dec = $$tableinfo{"declination"};
            $dec =~ s/^\@//;
            my $target_name = $$tableinfo{"target_name"};
            $target_name =~ s/^\@//;
            my $names = $$self{'names'};

            foreach (@$names) {
                if ($_ eq $ra) {
                    if ($coordsys =~ /galactic/i) {
                        $_ = "_Lii";
                    } elsif ($coordsys =~ /ecliptic/i) {
                        $_ = "_Elon";
                    }
                } elsif ($_ eq $dec) {
                    if ($coordsys =~ /galactic/i) {
                        $_ = "_Bii";
                    } elsif ($coordsys =~ /ecliptic/i) {
                        $_ = "_Elat";
                    }
                }
            }
        }
	
	$self->printColumnHeader($$self{"names"});
        print "<tbody>\n";
	
	$$self{"intable"} = 1;
    }
    
    #my $class = ($row % 2 ? "datacell1" : "datacell2");
    my $class = ($row % 2 ? "thinbordercell1" : "thinbordercell2");
    my $skip = $$self{"skipfield"};
    my $line = $$data{"data"};

    my $trans = $$self{"transform"};
    
    my ($racol, $deccol) = ($$trans{"racols"},$$trans{"deccols"});
    my $savera = Browse::CGI::Encode($$line[$$racol[0]]) if $$racol[0];
    my $savedec = Browse::CGI::Encode($$line[$$deccol[0]]) if $$deccol[0];
    $savedec = "%2B".$savedec unless (!$savedec or $savedec =~ /^-|^%2b|^%2d/i);
    Browse::TableFormat::transformLine($line, $trans);
    
    # We defer the transformat of spaces into NBSPs to here
    # because this is the point that we know that
    # the data is going to be presented in HTML.
    my $col;
    foreach $col(@$racol,@$deccol) {
        $$line[$col] =~ s/ /\&nbsp\;/g;
        $$line[$col] =~ s/!!/<a href="$::w3bhelp#coorderror">!!<\/a>/g;
    }
    
    
    my $unique = $$self{"meta"}{"table"} . "|||";
    my $types = $$data{"types"};

    my $longs = $$trans{"longs"};
    my ($names) = $$self{"names"};
    
    # Check if our skip fields are for multi-table result.
    # We want to put double pipes at the end of the keys
    # for each table.
    my $n0 = substr($$names[0], 0, 5);
    if ($n0 !~ /w3\_[a-z]\_/i) {
        $n0 = "w3_";
    }
   

    # Get the biblink info needed for the services link
    my $bkeys = $$self{"meta"}{"tableinfo"}{"bibkeys"};
    my $bkeystring;
    my $i;
    for ($i=0; $i <= $#$line; $i += 1) {
       my $sep = '|' if length($bkeystring);
       $bkeystring .= $sep."$$names[$i]=$$line[$i]" if $bkeys =~ /\b$$names[$i]\b/; 
    }
#print "bkeystring=$bkeystring<br>\n" if $::debug;
    my @biblink = split(/ /,$$self{"meta"}{"tableinfo"}{"bibcodes"}{$bkeystring});
    @biblink=grep(/\w/,@biblink); # get rid of any nulls
    my $biblink = "bibcode=".join("&bibcode=",@biblink) if scalar(@biblink);
 
    my $dlink="";
    my $oldsep;
    my $i;
    for ($i=0; $i <= $#$line; $i += 1) {
#print "line=@$line<br>\n" if $::debug;
#print "names=@$names<br>\n" if $::debug;
#print "$$names[$i]=$$line[$i]<br>\n" if $::debug;
        if ($i < $skip) {
	
	    my $str='';
	    my $name='';
	
	    if ($$types[$i] == 6) {
	        # Ensure adequate precision for double values.
		# By default the string representation of doubles
		# isn't adequate.
		if (length($$line[$i]) > 0) {
		    $str .= sprintf("%30.20g", $$line[$i]);
		    $str =~ s/^ *//;
		    $str =~ s/ *$//;
		}
	    } else {
	        $str = $$line[$i];
	    }
            $name = $$names[$i];
            $name =~ s/^w3_//i;
            $name =~ s/^\w_// if $$self{"meta"}{"table"} =~ / x /;
#print "trimmed<br>\n" if $$self{"meta"}{"table"} =~ / x / and $::debug;
	    
	    # Handle nulls.
	    if (length($str) > 0) {
                $unique .= $name ."=". $str . '|';
	    } else {
	        $unique .= $name."=".' |';
	    }
	    
	    if ($$names[$i+1] !~ /^$n0/i) {
	        $unique .= "|";
		$n0 = substr($$names[$i+1], 0, length($n0));
	    }
	    
	} else {
	    if ($i == $skip and $skip > 0) {
#print "unique=$unique<br>\n" if $::debug;
                my $cunique  = Browse::CGI::Encode($unique);
                my $ccunique  = Browse::CGI::Encode($cunique);
                $dlink = "/$::cgi/w3hdprods.pl?files=Preview&Coordinates=$$self{coordsys}&Equinox=$$self{equinox}&CheckSize=1&showgifs=1&Target=$ccunique" if $$self{"meta"}{"dpsets"};
    
                my $begin = "<a target=\"moreinfo\" href=\"w3hdprods.pl?".
                   "files=P&Target=$ccunique&Coordinates=$$self{coordsys}&Equinox=$$self{equinox}\"><img src=\"$::images/mag5.gif\" ".
                   "height=16 width=16 border=0 alt=\"Display all params\">".
                   "</a>";
                if ($unique =~ /\// && $unique !~ / x /i) {
                    my ($table,$keyval) = split('\|\|\|', $unique, 2);
                    ($keyval) = $keyval =~ /=\s*(.+)\s*\|\|$/;  
                    my $key = $$self{meta}{tableinfo}{"query_key"};
                   $begin .= "<a target=\"moreinfo\" href=".
                   "\"$::vizier/viz-bin/VizieR-5?-source=$table&".
                   "$key=$keyval\"><img src=\"$::images/magviz.gif\" ".
                   "height=16 width=16 border=0 alt=\"Display all params (from VizieR)\">".
                   "</a>";
                }
                $begin .= "<input type='checkbox' name='Target' id='Target' value='$cunique'>";
                print "<tr><td class='$class'>$begin</td>\n";

                if ($unique !~ / x /i) {
                  print "<td class='$class'>\n"; #Services cell
                  print $self->servicelinks($savera,$savedec,$dlink,$biblink);
                  print "</td>";
                }

	    }

	    
	    # Skip long strings.
	    if ($longs  && grep($_ == $i, @$longs)) {
	        next;
	    }
	  
            $$line[$i] =~ s/\s+/\&nbsp\;/g; # don't want to break the line
            print "<td class='$class'>$$line[$i]";
            # make empty cells look right
            print "&nbsp;" unless $$line[$i] =~ /\S/;
            print "</td>";
	}
    }

    print "</tr>\n";
    
    for ($i=0; $i <= $#$longs; $i += 1) {
        my $span = $#$line - $skip - 1;
	if ($span <= 0) {
	    $span = 1;
	}
	
	
        print "<tr><td colspan=2><td colspan=$span>";
	
	my $text = $$line[$$longs[$i]];
	$text =~ s/\0/<br>\n/g;
        print "$text</td></tr>\n";
    }
    $$self{"rowcount"} += 1;
    
}

sub servicelinks {
    my ($self, $savera, $savedec, $dlink, $biblink) = @_;
    my $html="";

    $html .= "<a title='See bibliographic references for this row' target='services' href='http://adsabs.harvard.edu/cgi-bin/bib_query?$biblink'>B</a>&nbsp;" if $biblink;
#    $html .= "<a target='services' href='http://adsabs.harvard.edu/cgi-bin/abs_connect?$biblink'>B2</a>&nbsp;" if $biblink;
    $html .= "<a title='View SkyView ROSAT All-sky survey image for these coordinates' target='services' href='http://skyview.gsfc.nasa.gov/cgi-bin/nnskcall.pl?VCOORD=$savera,$savedec&SURVEY=RASS-Cnt+Broad'>R</a>&nbsp;" if $savera =~ /\d/;
    $html .= "<a title='Query NED for these coordinates' target='services' href='http://nedwww.ipac.caltech.edu/cgi-bin/nph-objsearch?search_type=Near+Position+Search&in_csys=Equatorial&in_equinox=J2000.0&lon=".$savera."d&lat=$savedec&radius=2.0&out_csys=Equatorial&out_equinox=J2000.0&obj_sort=Distance+to+search+center&zv_breaker=30000.0&list_limit=5&img_stamp=YES&z_constraint=Unconstrained&z_value1=&z_value2=&z_unit=z&ot_include=ANY&nmp_op=ANY'>N</a>&nbsp;" if $savera =~ /\d/;
    $html .= "<a title='Query SIMBAD for these coordinates' target='services' href='http://simbad.harvard.edu/sim-id.pl?Radius=2&Ident=$savera+$savedec'>S</a>&nbsp;" if $savera =~ /\d/;
    $html .= "<a title='Preview data products for this row' target='services' href='$dlink'>D</a>&nbsp;" if ($$self{"meta"}{"dpsets"} and $dlink);
    return $html;
} # end of servicelinks


# Process empty tables or error messages.

sub showEmpties {

    my ($self, $empties, $errs) = @_;
    $self->closeTable();
    if ($#$empties >= 0) {
	
	print "<hr><p> <strong>No matches for</strong><br><dl>\n";
	my $niltab;
	foreach $niltab(@$empties) {
            my $stable = $$niltab[0];
            $stable =~ s/$::local_origin//g;
	    print "<dd> <a name=\"$stable\">$$niltab[1] ($stable)</a>\n";
	}
	print "</dl>\n";
    }
    my $lastErr;
    if ($#$errs >= 0) {
        print "<hr><p> <strong>Search errors</strong><br><dl>\n";
	my $errtab;
	foreach $errtab(@$errs) {
            my $stable = $$errtab[1];
            $stable =~ s/$::local_origin//g;
	    if ($$errtab[0] ne $lastErr) {
	        print "<dt> $$errtab[0] ";
	    }
	    print "<dd> <a name=\"$stable\">$$errtab[2] ($stable)</a>\n";
	}
	print "</dl>\n";
    }
}
        
# This is private method that sets up the
# table transformation to be used to format
# the table for output.

sub setTransform {

    my ($self, $data) = @_;

    my $names =  $$data{"names"};
print "names=@$names<br>\n" if $::debug;
    my $types =  $$data{"types"};
    
    my $coordsys  = $$self{"coordsys"};
    
    my $meta      = $$self{"meta"};
    my $tableinfo = $$meta{"tableinfo"};
    
    $$self{"unitFlds"} = undef;
    
    my @radecs;
    
    my $rac  = $$tableinfo{"right_ascension"};
    my $decc = $$tableinfo{"declination"};
    
    my (@raf)  = split('\|', $rac);
    my (@decf) = split('\|', $decc);
    
    if (defined($rac) && defined($decc)) {
        my $i;
        for ($i=0; $i <= $#raf; $i += 1) {
            $rac = $raf[$i];
            $decc = $decf[$i];

            if (substr($rac, 0, 1) eq "@") {
    	        $rac = substr($rac, 1);
    	    }
            if (substr($decc, 0, 1) eq "@") {
	        $decc = substr($decc, 1);
	    }

            push(@radecs,"$rac,$decc");
        }
    }
    
    my $paraminfo = $$meta{"paraminfo"};
    my ($accuracy) = $$paraminfo{$rac}{"parameter_format"} =~ /:\.(\d+)f/; 
    
    my @mjds;
    my $name;
    
    my @unitFlds = ();
    my $unitCount = 0;

    my @truncs;
    foreach $name (@$names) {


        my ($unit) = $$paraminfo{$name}{"parameter_unit"};
	$unit =~ s/^ *//;
	if (length($unit) > 0) {
            if ($unit eq "mjd") {
	        push(@mjds, $name);
		push(@unitFlds, "");
		
	    } elsif (substr($unit,0,3) ne "deg") {
	        push(@unitFlds, "[$unit]");
		$unitCount += 1;
	    } else {
	        push(@unitFlds, "");
	    }
	#} elsif ($name  =~ /\_.*offset/i) {
	#    # Won't display units if this is the only one, do don't increment unitCount
	#        push(@unitFlds, "[']");
	} else {
	    push(@unitFlds, "");
	}
	
	
	#if ($name =~ /\_.*(offset|delta)/i) {
	#    push(@truncs, "$name,\%0.3f");
	#}
	
	
	my ($form) = $$paraminfo{$name}{"parameter_format"};
	my ($bas, $out) = split('\:', $form);
	#if (defined($out)) {  #lmm
	if (defined($out) and $out =~ /\d/) {
	    push(@truncs, "$name,\%".$out);
	}
	    
    }
    
    if ($unitCount > 0) {
        $$self{"unitFlds"} = \@unitFlds;
    }
    
    # Just collect up the constants that we'll use in the
    # transformation of each line into a single array.
    my ($trans) = Browse::TableFormat::setTransform($names, $types, \@radecs, \@mjds, \@truncs,
	             $$self{"coordsys"});
		     
    $$trans{"equinox"}     = $$self{"equinox"};
    $$trans{"dateformat"}  = $$self{"dateformat"};
    $$trans{"coordformat"} = $$self{"coordformat"};
    $$self{"names"}        = $$trans{"names"};
    $$trans{"accuracy"} = $accuracy if $accuracy =~ /^\d+$/;
    
    return $trans;
}

# This prints out the columns at the beginning of each HTML table block.

sub printColumnHeader {

    my ($self, $names) = @_;
    
    my $meta  = $$self{"meta"};
    my $trans = $$self{"transform"};
    my $longs = $$trans{"longs"};

    print "<div class='lscrolltable'>" unless $$self{"intable"};
    $$self{"skipfield"} = 
      Browse::Results::PrintTableParams($$meta{"table"}, $$meta{"tableinfo"}, 
                                $$meta{"sortable"}, $names, $$self{"unitFlds"}, $longs,$$self{"offset_var"});
    my $fld;

    # Handle long strings
    foreach $fld(@$longs) {
        my $span = $#$names - $$self{"skipfield"} - 1;
	
        if ($span <= 0) {
            $span = 1;
        }

        print "<tr><th class='thinborderhead' colspan='2'></th><th class='secondheadrow' scope='col' colspan='$span' align='left'>";
	my $fn = $$names[$fld];
	my $doc = $$meta{"tableinfo"}{"table_document_url"};
	print "<a target='helpwin' href='$doc#$fn'> $fn </a>";
	print "</th></tr>\n";
    }
        print "</thead>\n";	
  
}

# Finish up processing of a table.  Close any open HTML blocks....

sub closeTable {

    my ($self, $msg) = @_;
    
    if ($$self{"intable"}) {
        print "</tbody>\n";
        print "</table></div>\n";
	$$self{"intable"} = 0;
        my $s = "";
        if ($$self{"rowcount"} > 1) {
            $s = "s";
        }
        my $table = $$self{"meta"}{"table"};
        my $xtable = $self->chopTable($table);

        print "$$self{rowcount} row$s retrieved from $xtable<p>\n";
        print "$msg<br>\n" if $msg;
	
	my $meta = $$self{"meta"};
	my $dp   = $$meta{"dpsets"};
	if (ref($dp)  && scalar keys %$dp == 0) {
	    $dp = undef;
	}
	

	my $tableinfo = $$meta{"tableinfo"};
        print "<table border='1'><tr>\n";
	if ($dp) {
            print "<td valign=\"top\">\n";
            $$tableinfo{table_name} = $table;
	    Browse::Results::DisplaySets($dp, $tableinfo);
            print "</td>\n";
	}
        print "<td valign=top>\n";
        my $coords = ($$tableinfo{right_ascension}? 1:0);
	Browse::Results::DisplayOptions($table,$coords,$dp);
        print "</td></tr></table>\n";
	#Browse::Results::DisplayButtons();

    }
}

sub YourSearchWas {
    my ($self,$positions,$qualifiers,$sorts,$dirs,$resmsg) = @_;
    use Browse::TableFormat;
    my $maxrows=$self->MaxRows;
    my ($coord_sys,$equinox) = $self->getCoordSys();

    my $f = "<span class=\"small\">";
    my $fe = "</span>";

    #These are wrong if there are more than one sort param - fix
    my ($sortinfo,$sort);
    foreach $sort (@$sorts) {
        $sortinfo .= $sort;
    }
    my ($dirinfo,$dir);
    foreach $dir (@$dirs) {
        $dirinfo .= $dir;
    }

    my $radius;
    my($num, $obj,$pos,$object_info);
    foreach $pos (@$positions){
       my ($coord1,$coord2,$rad,$name,$resolver) = split('\|',$pos);
       $radius = $rad;

       my $format = "decimal" if $coord_sys =~ /gal/i;
       ($coord1,$coord2)  = Browse::TableFormat::fromJ2000($coord1,$coord2,$coord_sys,$equinox,$format);
       $coord1 .= "&#176;" if $format =~ /decimal/;
       $coord2 .= "&#176;" if $format =~ /decimal/;
       $coord1 =~ s/\s+/&nbsp;/g;
       $coord2 =~ s/\s+/&nbsp;/g;

       $object_info .= " </tr> <tr><td>&nbsp;</td>\n" if $num;

       if ($name =~ /[^\d\.\+\-\s\,]/) {
            $obj = "<td>$f <i>$name,</i> resolved by $resolver " .
                  "to [&nbsp;$coord1,&nbsp;$coord2&nbsp;] $fe</td>";
       }else{
            $obj = "<td>$f $name $fe</td>";
       }
       $object_info .= "    ".$obj . "\n";
       $num++;
    }

    my ($num2,$qual,$qual_info);
    foreach $qual (@$qualifiers){
       my ($field,$op,$min,$max) = split('\|',$qual);
       $min = DateTime::mjd2iso($min) if defined $min and $field =~ /metaparam_start_time|metaparam_end_time/;
       $max = DateTime::mjd2iso($max) if defined $max and $field =~ /metaparam_start_time|metaparam_end_time/;
       $field = "<i>$1</i>" if $field =~ /_metaparam_(.*)/;
       $qual_info .= " AND<br>" if $num2;
       if ($op =~ /between/) {
          (my $not, $op) = $op =~ /(!?)(.*)$/;
          $not = "not" if $not;
          if (!$min or $min =~ /'\s?'/) {
             if (!$not) {
                 $qual_info .= "$field <= $max";
             } else {
                 $qual_info .= "$field > $max";
             }
          } elsif (!$max or $max =~ /'\s?'/) {
             if (!$not) {
                 $qual_info .= "$field >= $min";
             } else {
                 $qual_info .= "$field < $min";
             }
          } else {
             $qual_info .= "$field $not $op $min and $max";
          }
       } else {
          $qual_info .= "$field $op $min";
       }

       $qual_info  =~ s/\s+/&nbsp;/g;
       $num2++;

    }

    if ($radius =~ /\d/) {
        my $radunit = "&deg;";
        if ($radius < 1) {
           $radius *= 60;
           $radunit = "\'";
        }
        if ($radius < 1) {
           $radius *= 60;
           $radunit = "\"";
        }
        $radius = sprintf("%.2f", $radius) . " $radunit";
    }

    #print search criteria
    my $string = "";
    $string .= "
        <br><center>
        <table><tr><td>
               <table class='thinbordertable'><tr>
        <td colspan='2' align='left'><b>Search was based on: </b></td></tr>";

    if ($object_info) {
       $string .= "<tr><td>$f Object/Coordinates: $fe</td>$object_info  </tr>";
    }
    if ($resmsg) {
       $string .= "<tr><td colspan='2'><span class='note'>$resmsg</span></td></tr>";
    }

    $string .= "<tr><td>$f Coord. System: $fe</td><td>$f $coord_sys";
    if ($coord_sys !~ /Galactic/i) {
         $string .= ", equinox  $equinox";
    }
    $string .= "$fe </td></tr>\n";

    $maxrows = "(no limit)" unless $maxrows;
    $string .= "<tr><td>$f Maximum Rows: $fe</td><td>$f $maxrows $fe</td></tr>\n";
    if ($qual_info) {
       $string .= "<tr><td>$f Parameter Query: $fe</td><td>$f $qual_info $fe</td></tr>";
    }

    if ($object_info) {
       $string .= "<tr><td>$f Search Radius: $fe</td><td>$f $radius $fe</td></tr>\n";
    }

    $dirinfo = "" if $dirinfo =~ /down/ or !$dirinfo;
    $dirinfo = ", reverse order" if $dirinfo =~ /up/;
    $string .= "<tr><td>$f Sorted by: $fe</td><td>$f ${sortinfo}$dirinfo $fe</td></tr>" if $sortinfo;

    $string .="</table></td><td><table>";
    $string .="<tr><td align='right'><input type='submit' name='redisplay' value='Redisplay'></td><td>as <select name='displaymode'>";
    $string .= "<option value='Display'>HTML Table" unless ref($self) =~ /::Display/;
    $string .= "<option value='TextDisplay'>Text Table" unless ref($self) =~ /::Text/;
#    $string .= "<option value='NonTableDisplay'>Faster HTML display" unless ref($self) =~ /::NonTable/;
    $string .= "<option value='FitsDisplay'>FITS Table" unless ref($self) =~ /::Fits/;
    $string .= "<option value='XMLDisplay'>astrores XML" unless ref($self) =~ /::XML/;
    $string .= "<option value='VODisplay'>VOTable XML" unless ref($self) =~ /::VO/;
    $string .= "<option value='ExcelDisplay'>Excel-compatible" unless ref($self) =~ /::ExcelDisplay$/;
#    $string .= "<option value='ExcelDisplay2'>Excel-compatible (mod_perl test)" unless ref($self) =~ /::ExcelDisplay2/;
    $string .= "</select></td></tr>";
#    $string .="<tr><td align='right'><input type='button' name='hera' value='Start Hera' onClick='window.open(\"/cgi-bin/hera/heraclient.pl\",\"herawin2\")></td><td><a target='helpwin' href=\"/hera\">What is Hera?</a></td></tr>";
    $string .="<tr><td align='right'><input type='reset'></td></tr>";
    $string .= "</table></td></tr>";

    $string .= "</table></center>\n";

print $string;

}


## Successful load
1;

# $Log$
