# $Id: QueryRequest.pm,v 1.12 2007/03/26 19:35:28 lmm Exp $
# $Log: QueryRequest.pm,v $
# Revision 1.12  2007/03/26 19:35:28  lmm
# Added target name to Search Offset column
#
# Revision 1.11  2006/06/23 17:21:35  lmm
# Setting offset to "_Search_Offset" if it has not been defined.
# Added method to return table sort.
#
# Revision 1.10  2003/07/31 14:09:27  cwh
# Added the reformating of the radius to print the correct accuracy if the
#   source is also printed.
#
# Revision 1.9  2003/07/23 20:15:10  cwh
# Added the printing of the user's input source in the radius field if the
# user asked for more than one source.
#
# Revision 1.8  2001/12/06 16:49:51  cwh
# Added use CachedQuery and use CrossRequest.
# Added setOffsetVar and getOffsetVar methods to isolate offset variable name.
#
# Revision 1.7  2001/11/07 23:34:19  cwh
# Moved everything except DBD/ into lib/Browse/, so as not to pollute the module
# namespace with our Sort, Search, CGI, etc.  This required putting Browse:: in
# front of every module call and package declaration.  Also any mod::new() had
# to be changed to mod->new().
#
# Revision 1.6  2001/06/20 22:57:34  cwh
# Got rid of $::writeHeader in calls to ::Error.
#
# Revision 1.5  2001/06/08 19:46:38  cwh
# Now passing $::writeHeader to ::Error calls, so we don't get repeating headers.
#
# Revision 1.4  2001/05/22 17:52:03  cwh
# Added Id and Log to files without them.
#

use strict;

# This package defines a superclass of database queries.
# Basically the user instantiates a query request and
# then calls getResult until a null is returned.  Each
# getResult will return either a metadata parcel describing
# subsequent results, or a data parcel describing a single row
# of results.  The metadata parcel is a hash with a variety of
# fields, notably a 'tableinfo' hash giving information about
# the table, and a 'paraminfo' hash giving information about
# the fields that are returned in the query.
# The data parcel is a hash with elements including the data
# values, column names, and column types as members.
#
# This package also includes a serialization/deserialization capability
# which may be used to 'pickle' scalars, hashes, arrays and objects.
#
# Instance fields:
#    table     -- a string indicating the table to be queried.
#    fields    -- an array reference of the field names to be queried.
#    positions -- an array reference of positions for which a cone
#                 search is desired.
#    qualifications -- an array reference of qualifications on 
#                      the query.
#    sorts     -- an array refernce of fields to sort by.
#    dirs      -- an array reference indicating that the
#                 corresponding field in sort is to be sorted
#                 in reverse order if defined.
#
#

package Browse::QueryRequest;

use Browse::HeasarcQuery;
use Browse::VizierQuery;
use Browse::CachedQuery;
use Browse::CrossRequest;

# Constructor
#
sub new {

    my ($self,$table) = @_;
    my $obj = {};
    my $key;
   
    $$obj{"table"} = $table;
    
    if ($table =~ /\//) {
        return bless $obj, "Browse::VizierQuery";

    } else {
        bless $obj, "Browse::HeasarcQuery";
        return $obj;

    }
}

# Start the query.
sub initiate {

    my ($self) = @_;
    $$self{"status"}       = "initiate";
}

# This routine transforms a string back to a Perl element
# which may be a scalar, array, hash or Object.  Only
# Hash objects are supported.

sub deserialize {

    my ($str, $offset) = @_;
    if (!defined($offset)) {
        $offset = 0;
    }

    my $p1 = substr($str, $offset, 1);

    if ($p1 eq "{") {
        my $fld = {};
        return deserHash($fld, $str, $offset+1);

    } elsif ($p1 eq "(") {
        my $fld =[];
        return deserArr($fld, $str, $offset+1);

    } elsif ($p1 eq "*") {
        my $comma = index($str, ",", $offset+1);
        if ($comma < 0) {
            Browse::Utils::Error("Invalid object reconstruction: No comma after name");
        }
        my $name = Browse::CGI::Decode(substr($str, $offset, $comma-$offset));
        my $fld = {};
        bless $fld, $name;
        if (substr($str, $comma+1, 1) ne "{") {
            Browse::Utils::Error("Invalid object reconstruction: Object hash");
        }

        return deserHash($fld, $str, $comma+2);

    } elsif ($p1 eq "'") {
        my $quote = index($str, "'", $offset+1);
        my $val = Browse::CGI::Decode(substr($str, $offset+1, $quote-$offset-1));
        return ($val, $quote+1);

    } elsif (substr($str, $offset, 4) eq "null") {
        return(undef, $offset+4);

    } else {
        Browse::Utils::Error("Invalid deserialization string $str at: $offset\n");
    }

}

# Deserialize a Perl ARRAY
sub deserArr {

    my ($arr, $str, $offset) = @_;

    my $val;

    while (substr($str, $offset, 1) ne ")") {
        ($val, $offset) = deserialize($str, $offset);
        push(@$arr, $val);
        if (substr($str, $offset, 1) eq ",") {
            $offset += 1;
            next;
        }
        if (substr($str, $offset, 1) eq ")") {
            last;
        }
        Browse::Utils::Error("Invalid character in array: $offset");
    }
    return ($arr, $offset+1);
}

# Deserialize a Perl HASH
sub deserHash {
    my ($hash, $str, $offset)  = @_;
    my $val;

    while (substr($str, $offset, 1) ne "}") {
        my $comma = index($str, ",", $offset);
        if ($comma < 0) {
            Browse::Utils::Error("Missing key in Hash");
        }
        my $key = Browse::CGI::Decode(substr($str, $offset, $comma-$offset));
        $offset = $comma + 1;
        my $val;
        ($val, $offset) = deserialize($str, $offset);
        $$hash{$key} = $val;
        if (substr($str, $offset, 1) eq ",") {
            $offset += 1;
            next;
        }
        if (substr($str, $offset, 1) eq "}") {
            last;
        }
        Browse::Utils::Error("Invalid character in array: $offset");
    }
    return ($hash, $offset+1);
}

# Create a string representing the state of a Perl element.
# The element may be a scalar, array, hash or object.
# We only handle objects created from blessed hashes.
# Circular references will cause an infinite
# loop.  The following syntax is used:
#    null                        for undefined values
#    'CGI-encoded value'         for scalars.
#    (val,val,val,...)           for arrays
#    {key,val,key,val...}        for hashes
#    *class,{key,val,key,val...} for objects
# Since all scalars are CGI-encoded, the parsing can
# assume punctuation characters are structural.
# No special order is maintained for hash keys.
sub serialize {
    my ($val) = @_;
    my $str;
    if (!defined($val)) {
        $str = "null";
    } elsif (!ref($val)) {
        $str = "'".Browse::CGI::Encode($val)."'";
    } elsif (ref($val) eq "ARRAY") {
        $str = "(";
        my $comma;
        foreach (@$val) {
            $str .=$comma . serialize($_);
            $comma = ",";
        }
       $str .= ")";

    } else {
        if (ref($val) ne "HASH") {
            $str = "*".Browse::CGI::Encode(ref($val)).",";
        }
        $str .= "{";
        my $comma;
        foreach (keys(%$val)) {
            $str .= $comma.Browse::CGI::Encode($_).','.serialize($$val{$_});
            $comma = ",";
        }
        $str .= "}";
    }
}

# Create a request based upon the stored state
# or just the table name if no state is specified.

sub regenerate  {

    my ($table, $state) = @_;
    my $obj;
    if ($state) {
        ($obj) = deserialize($state);
    } else {
        $obj = new($table);
        $obj->setTable($table);
    }

    return $obj;
}

# Save the current state of the request in the
# request.

sub storeState {
    my ($self) = @_;
    my $serialization = serialize($self);
    $$self{"serialization"} = $serialization;
}

# Copy constructor
sub copy {

    my ($self) = @_;
    my $copy = {};
    
    my $key;
    foreach $key (%$self) {
        $$copy{$key} = $$self{$key};
    }
    
    bless $copy, ref($self);
    return $copy;
}

# Accessor functions.

# Table -- the name of the table to be queried.
sub setTable {
    my ($self, $table) = @_;
    $$self{"table"} = $table;
}

sub getTable {
    my ($self) = @_;
    return $$self{"table"};
}


# The fields to be returned.

sub setFields {
    my ($self, $fields) = @_;
    $$self{"fields"} = $fields;
}

sub getFields {
    my ($self) = @_;
    return $$self{"fields"};
}

sub addField {
    my ($self, $field) = @_;
    
    my $fields = $$self{"fields"};
    push(@$fields, $field);
}

# The name of the offset variable in the query
sub setOffsetVar {
    my ($self,$offset) = @_;
    $$self{"offset_var"} = $offset || "_Search_Offset";
}

sub getOffsetVar {
    my ($self) = @_;
    return $$self{"offset_var"} || "_Search_Offset";
}

# Maximum number of rows to display.
sub setMaxRows {
    my ($self, $max) = @_;
    $$self{"maxrows"} = $max;
}

sub getMaxRows {
    my ($self) = @_;
    return $$self{"maxrows"};
}

# The table prefix associated with the first table in this query.

sub setPrefix {
    my ($self, $prefix) = @_;
    $$self{"prefix"} = $prefix;
}

sub getPrefix {
    my ($self) = @_;
    return $$self{"prefix"};
}

# Get the next prefix available.
sub nextPrefix {
    my ($self) = @_;
    my $prefix = $$self{"prefix"};
    if ($prefix) {
        $prefix =~ tr/abcdefghijklmnopqrstuvwxyz/bcdefghijklmnopqrstuvwxyzA/;
    }
    return $prefix;
}
  

# Postions around which to do a cone search.
sub setPositions {
    my ($self, $positions) = @_;
    $$self{"positions"} = $positions;
}

sub getPositions {
    my ($self) = @_;
    return $$self{"positions"};
}

sub addPosition {
    my ($self, $position) = @_;
    
    my $positions = $$self{"positions"};
    
    if (!defined($positions)) {
        $positions = [];
	$$self{"positions"} = $positions;
    }
    
    push(@$positions, $position);
}

# Table -- the name of the table to be queried.
sub setDefaultRadius {
    my ($self, $rad) = @_;
    
    if (substr($rad, 0, 1) eq '@') {
        $rad = substr($rad, 1);
    }
    $$self{"radius"} = $rad;
}

# The default search radius
sub getDefaultRadius {
    my ($self) = @_;
    return $$self{"radius"};
}

# Non cone-search qualifiers.

sub setQuals {
    my ($self, $quals) = @_;
    $$self{"quals"} = $quals;
}

sub getQuals {
    my ($self) = @_;
    return $$self{"quals"};
}

sub addQual {
    my ($self, $qual) = @_;
    
    my $quals = $$self{"quals"};
    if (!defined($quals)) {
        $quals = [];
	$$self{"quals"} = $quals;
    }
    push(@$quals, $qual);
}

# Sort fields.

sub setSort {
    my ($self, $sorts, $dirs) = @_;
    $$self{"sorts"} = $sorts;
    $$self{"dirs"}  = $dirs;
}

sub getSort {
    my ($self) = @_;
    return ($$self{"sorts"}, $$self{"dirs"});
}
sub getSortString {
    my ($self) = @_;
    return $$self{"sortString"};
}

sub addSort {

    my ($self, $field, $dir) = @_;
    
    my $sorts = $$self{"sorts"};
    if (!defined($sorts)) {
        $sorts = [];
	my $dirs = [];
        $$self{"sorts"} = $sorts;
	$$self{"dirs"} = $dirs;
    }
    
    push(@$sorts, $field);
    if (defined($dir)) {
        my $dirs = $$self{"dirs"};
	$$dirs[$#$sorts] = $dir;
    }
}

# Is this a local query.  By default we assume not.
sub localQuery {
    return 0;
}

# get information on the offset column. Used by getResult below.
sub getOffsetInfo {
    my ($self,$data) = @_;
    my $offsetvar=$self->getOffsetVar();
    ($$self{"offsetaccuracy"}) = 
       $$self{"paraminfo"}{$offsetvar}{"parameter_format"} =~ /:\.(\d+)f/ 
       unless defined $$self{"offsetaccuracy"};
    $$self{"offsetaccuracy"} ||= "3"; # default 3 decimal places
    $$self{"offsetvar"} = $offsetvar;
    ## unset the formatting so the source doesn't get truncd
    #$$self{"paraminfo"}{$offsetvar}{"parameter_format"} ='';

    my $count=0;
    my $names = $$data{"fields"};
    foreach my $n (@$names) {
        if ($n =~ /^$offsetvar$/) {
            $$self{offsetcol} = $count;
            last;
        }
        $count++;
    }
}

# Get the next element of data.
# Note that it might be either a line of results
# or metadata.  We are guaranteed to get
# metadata returned before the first line.
# but there may be multiple sets of metadata.
#
# This method handles multiple positional
# queries by querying the underlying database
# multiple times once for each position.
# It assumes the metadata returned such queries
# is the same and discards the metadata returned
# after the first query.


sub getResult {

    my ($self) = @_;
    my $data;
    

    if ($$self{"status"} eq "initiate") {
        #--- first time through
        my $pos = $$self{"positions"};
	
	# Handle query of multiple positions
	# as a sequence of queries of a single position.
	
	if ($#$pos > 0) {
            #--- multiple positions
            my $copy = $self->copy();
	    $$self{"currentQuery"} = $copy;
	    $$copy{"allpositions"} = $$self{"positions"};
	    
	    my $cpos;
	    
	    for ($cpos=0; $cpos <= $#$pos; $cpos += 1) {
	        $copy->setPositions([$$pos[$cpos]]);
		$data = $copy->getData();
                $copy->getOffsetInfo($data);
		if ($data) {
		    last;
		}
		$copy->qclose();
		
	    }
	    
	    $$self{"positionIndex"} = $cpos;
	    $$self{"status"} = "data";
	    $$self{"currentTable"} = $$copy{"currentTable"};
	    
	} else {
	    $data = $self->getData();
            $self->getOffsetInfo($data);
	    $$self{"status"} = "data";
	
	}
	
    } else {
    
        my $curr = $$self{"currentQuery"};
        if ($curr) {
            #--- Set offset column content for multiple positions
	    my $cpos = $$self{"positionIndex"};
	    my $pos  = $$self{"positions"};
	    
            while (1) {
	    
	        $data = $curr->getData();
		if ($data) {
		    if ($$data{"type"} ne "meta") {
                       # insert the 'from' source
                       (my $entry) = (split('\|',$$pos[$cpos]))[3];
                       my $accuracy = $$curr{"offsetaccuracy"};
                       my $offsetcol = $$curr{"offsetcol"};
                       my $offsetvar = $$curr{"offsetvar"};
                       $$data{"data"}[$offsetcol]  =  
                         sprintf("%.${accuracy}f", $$data{"data"}[$offsetcol]) .
                         " ($entry)" if defined $$data{"data"}[$offsetcol];

                       # unset the formatting so the source doesn't get truncd
                       $$curr{"paraminfo"}{$offsetvar}{"parameter_format"} ='';
		       last;
		    } else {

		        my $mtable = $$data{"table"};
			if ($$self{"currentTable"} eq $mtable) {
			    $$self{"status"} = "data";
		            next;
			} else {
			    $$self{"currentTable"} = $mtable;
			    last;
			}
		    }
		}
		
		if ($cpos >= $#$pos) {
		    last;
		}
		$cpos += 1;
		$curr->qclose();
		
		$$self{"positionIndex"} = $cpos;
		$curr->setPositions([$$pos[$cpos]]);
	    }
	    
	} else {
            #--- Set offset column content for only one target position 
	    my $pos  = $$self{"positions"};
            (my $entry) = (split('\|',$$pos[0]))[3];
	    $data = $self->getData();
            if ($data) {
               $self->getOffsetInfo($data);
               # insert the 'from' source
               my $accuracy = $$self{"offsetaccuracy"};
               my $offsetcol = $$self{"offsetcol"};
               my $offsetvar = $$self{"offsetvar"};
               $$data{"data"}[$offsetcol]  =  
                  sprintf("%.${accuracy}f", $$data{"data"}[$offsetcol]) .
                  " ($entry)" if (defined $$data{"data"}[$offsetcol] && 
			$entry =~ /\w/);

                  ## unset the formatting so the source doesn't get truncd
               $$self{"paraminfo"}{$offsetvar}{"parameter_format"} ='';
            }
	}
    }
    
    return $data;
}

sub qclose {

    my ($self) = @_;
    $$self{"status"} = "initiate";
}

1;
