# $Id: TextDisplay.pm,v 1.25 2008/12/15 19:28:28 lmm Exp $
# $Log: TextDisplay.pm,v $
# Revision 1.25  2008/12/15 19:28:28  lmm
# Made changes to accommodate multiple coordinate pairs in heasarc tables.
#   These new coordinates are treated the same as standard ra and dec.
#
# Revision 1.24  2007/10/09 14:10:50  lmm
# Added empty methods beforeNewTable() and afterTable() .
#
# Revision 1.23  2007/03/28 15:30:27  lmm
# Added calls to HTML::Encode() where table column value may contain < or >
#   or other symbol that should not be interpreted as HTML.
#
# Revision 1.22  2007/03/26 19:14:10  lmm
# Data products display options are not displayed if there are none available
#   for the rows in results tables.
#
# Revision 1.21  2006/06/23 18:47:51  lmm
# Made change to Results routine call since Results is now a class
#
# Revision 1.20  2006/01/31 20:45:01  lmm
# Added hidden form input element with serialized query string value to
#   pageHeader() to get Save Query To File function to work
#
# Revision 1.19  2006/01/12 17:29:58  lmm
# Changed call to Display::resultsSummary() to reflect new argument
#
# Revision 1.18  2005/12/16 15:30:59  lmm
# Major overhaul of Text output format - search based on information,
#   data products selections, plotting and cross-cor options, results summary
#
# Revision 1.17  2005/11/25 21:39:06  lmm
# Fixed bug that was causing the column data types to be incorrect.  Any
#   field removed from column list must also be removed from type list.
#   (affected Fits Display)
#
# Revision 1.16  2005/11/21 20:19:03  lmm
# Added more advanced features (data products, pplot, cross correlation) to
#   make Text Table output format  more usable
#
# Revision 1.15  2005/09/12 20:22:54  lmm
# Corrected elimination of fields used for related links that are not to be
#   displayed in the data rows.
#
# Revision 1.14  2005/03/09 14:03:06  lmm
# Replaced deleted call to TableFormat::transformLine()
#
# Revision 1.13  2005/02/15 15:59:40  lmm
# Skip non default display set parameters (_link_ prefix) that are used
#    for table and bibliography links.
# Overrided resultsSummary method with empty method
#
# Revision 1.12  2003/09/03 21:12:31  lmm
# Added empty YourSearchWas method to override main Display class method
#
# Revision 1.11  2003/08/29 15:19:49  lmm
# Defined text for page header
# Added missing Coordinate system and Equinox to Search Based on/Requery
#    box
#
# Revision 1.10  2003/07/31 19:10:38  cwh
# Overriding legend subroutine.
#
# Revision 1.9  2003/07/23 20:44:03  cwh
# Fixed display so that results are presented in fixed-width font. (Just 'pre'
# wasn't working.)
#
# Revision 1.8  2003/04/28 01:47:01  cwh
# Added delimiter specification option.
#
# Revision 1.7  2002/12/16 22:19:38  cwh
# Added DisplayLinks so that the redisplay will work from the text display.
#
# Revision 1.6  2002/10/17 18:34:06  cwh
# removed plotsql from vars passed.
#
# Revision 1.5  2002/05/14 18:05:56  cwh
# Added empty Directions() and showText().
#
# Revision 1.4  2002/03/17 19:11:38  cwh
# Now printing regular Browse Header.
# Added ra/dec -> lii/bii label conversion for galactic coords
#
# Revision 1.3  2001/11/20 15:47:43  cwh
# Made looking for "w3_" be case-insensitive.
#
# Revision 1.2  2001/11/07 23:34:21  cwh
# Moved everything except DBD/ into lib/Browse/, so as not to pollute the module
# namespace with our Sort, Search, CGI, etc.  This required putting Browse:: in
# front of every module call and package declaration.  Also any mod::new() had
# to be changed to mod->new().
#
# Revision 1.1  2001/10/25 20:12:14  cwh
# Separated these from the Display.pm module.
#

# This class displays results in HTML where
# the individual lines are not included in tables.

use strict;

package Browse::TextDisplay;
use Browse::SearchBasedOn;
use HTML;

use vars qw( @ISA );
@ISA = qw( Browse::Display );

# Minimize information at the top of the page.
sub pageHeader {

    my ($self, $navtrail) = @_;

    # Add header and form/action tags
    Browse::Results::Header(0,0,"Text Display of Query Results");

    print "<script type=\"text/javascript\" language=\"Javascript\">\n";
    Browse::ValidateInput::tickTable();
    print "</script>\n";

    print "<strong> HEASARC Table Query: Text Results</strong><p>\n";

    # values necessary for Requery 
    print "<input type='hidden' name='Coordinates' value='".
	$$self{"coordsys"}."'>\n";
    print "<input type='hidden' name='Equinox' value='".
	$$self{"equinox"}."'>\n";
    print "<input type='hidden' name='savedQuery' value='".
	$$self{"savedquery"}."'>\n";

    $self->{"dpavailable"}=0;  #-- flag for data products availabilty
}

# Can't print rows until we've got them all since
# we need to 'neaten' the table output and make all rows
# have columns of the same width.  So transform
# the fields but just save the results for later printing.

sub processData {
    
    my ($self, $data) = @_;
    
    my $row = $$self{"rowcount"};
    my $xn = $$data{"names"};
    if (!$$self{"transform"}) {
    
        $$self{"transform"}  = $self->setTransform($data);
        $$self{"dataarray"} = [];
        $$self{"intable"} = 1;
	$$self{"names"} = $$data{"names"};
	$$self{"types"} = $$data{"types"};
    }
    
    # Copy the line data -- the data array may be reused.
    my $line  = $$data{"data"};
    my $type  = $$data{"types"};
    
    my @line;
    my @validtypes; 
    
    ##--- look for parameters not in display set and remove
    my @keep;
    my @names;

    #--- look for prefix indicating columns that are not-to-be-displayed. 
    #--- save names and indexes of columns to be displayed 
    for (my $i=0; $i < scalar(@$xn); $i++) {
       #print "TextDisp:$$xn[$i]<br>" if $::debug;
       if ($$xn[$i] !~ /^_link_/) {
           
          push (@names,$$xn[$i]);
          push (@keep,$i);
       } 
    }
        
    #--- go through the row data and remove columns and types that should not 
    #--- be displayed.  (columns and types should match for indexing)
    for (my $j=0; $j <scalar(@$line);$j++) {
       if (scalar(grep /$j/, @keep)) {
          #--  save column if index is in "keep column" array
          push (@line,$$line[$j]);
          push (@validtypes,$$type[$j])
       } 
    }
    	
    my $trans = $$self{"transform"};
    my $arr   = $$self{"dataarray"};
    
    
    Browse::TableFormat::transformLine(\@line, $trans);

    
    push(@$arr, \@line);  #--- save row data
    $$self{"rowcount"} += 1;
    $$self{"names"} = \@names;   # update var with column names to be displayed
    $$self{"types"} = \@validtypes;   # update var with column names to be displayed
}

# For this class, closeTable actually
# prints the results after 'neatening' the table.

sub closeTable {

    my ($self) = @_;
    
    if ($$self{"intable"}) {
	$$self{"intable"} = 0;
	
	my $array = $$self{"dataarray"};
	my $meta  = $$self{"meta"};
	my $names = $$self{"names"};
	my $types = $$self{"types"};
        my $table = $$self{"meta"}{"table"};
        my $shortname = $self->chopTable($table);

    
        my $nskip = 0;
        while ($nskip <= $#$names) {
            
	    if (substr($$names[$nskip], 0, 3) !~ /w3_/i) {
	            last;
	    } 
	        $nskip += 1;
        }
    
        my $coordsys  = $$self{coordsys};
        my $tableinfo = $$meta{"tableinfo"};
    
        my $ra = $$tableinfo{"right_ascension"};
        $ra =~ s/^\@//;
        my $dec = $$tableinfo{"declination"};
        $dec =~ s/^\@//;
    
        #--- Change table header for non-equatorial coordinates
        #my ($oras, $odecs) = $self->otherRasDecs($tableinfo);
        my ($oras, $odecs) = Browse::Display::otherRaDecs($tableinfo);
   
        foreach (@$names) {

           $_ = Browse::Display::renameRaDecs($_, $ra, $dec, $oras, $odecs, 
		$coordsys );

           #if (scalar(@$oras) && scalar(@$odecs)) {
              #my @check_other_ras = grep /^$_$/, @$oras;
              #my @check_other_decs = grep /^$_$/, @$odecs;
              #if ($coordsys =~ /galactic/i ) {
                 #if (scalar(@check_other_ras)) {
                    #$_ =~ s/ra$/lii/;
                 #}
                 #if (scalar(@check_other_decs)) {
                    #$_ =~ s/dec$/bii/;
                 #}
              #} elsif ($coordsys =~ /ecliptic/i) {
                 #if (scalar(@check_other_ras)) {
                    #$_ =~ s/ra$/elon/;
                 #}
                 #if (scalar(@check_other_decs)) {
                    #$_ =~ s/dec$/elat/;
                 #}
              #}
           #}


           #if ($_ eq $ra ) {
              #if ($coordsys =~ /galactic/i) {
                 #$_= "_Lii";
              #} elsif ($coordsys =~ /ecliptic/i) {
                 #$_ = "_Elon";
              #}
           #} elsif ($_ eq $dec) {
              #if ($coordsys =~ /galactic/i) {
                 #$_ = "_Bii";
              #} elsif ($coordsys =~ /ecliptic/i) {
                 #$_ = "_Elat";
              #}
            #}
        }
    
    	
        Browse::TableFormat::neatenTable($array, $names, $types, undef);
    
    	
        my $trans = $$self{"transform"};
        my $longs = $$trans{"longs"};
        my $delim = $$self{"delimiter"};
        $delim = "|" unless defined $delim;
    	
        print "<p><pre><font face='courier'>\n";
            
        my $html;
        if ($nskip > 0) {
           # Leave a space for the check box  if rows have unique ids (nskip>0)
           my $selectAlltable = $table;    #
           $selectAlltable =~ s/$::local_origin//gi;
           ($selectAlltable) = split(" ", $selectAlltable);

            $html = "<script type='text/javascript' language='JavaScript1.2'>" .
		"<!-- \ndocument.write (\"<input type='checkbox' " .
		"id='allobs_$selectAlltable' name='allobs_$selectAlltable' " .
		"onClick=tickTable(this.form.Target,this.checked," .
		"'$selectAlltable')>" .
		"<label for='allobs_$selectAlltable'>Select All</label>\"); " .
		"// --> </script>\n ";

        }
        print $html;
        print "   ";


        #--- print column names
        $self->displayTextLine($names, $nskip, $longs,$delim);
    	
        my $line;
        my $unique = "";
        foreach $line(@$array) {
           #--- Generate data products link
           for (my $i; $i < $nskip; $i++) {
              $unique = $table . "|||";
              my $str='';
 	      my $name='';
	
              # Check if our skip fields are for multi-table result.
              # We want to put double pipes at the end of the keys
              # for each table.
              my $n0 = substr($$names[0], 0, 5);
              if ($n0 !~ /w3\_[a-z]\_/i) {
                  $n0 = "w3_";
              }
	      if ($$types[$i] == 6) {
	          # Ensure adequate precision for double values.
		  # By default the string representation of doubles
		  # isn't adequate.
		  if (length($$line[$i]) > 0) {
		      $str .= sprintf("%30.20g", $$line[$i]);
		  }
	      } else {
	          $str = $$line[$i];
	      }
	      $str =~ s/^ *//;  #- get rid of blank spaces
	      $str =~ s/ *$//;

              $name = $$names[$i];
              $name =~ s/^w3_//i;
              $name =~ s/^\w_// if $table =~ / x /;
  
  
	      # Handle nulls.
	      if (length($str) > 0) {
                  $unique .= $name ."=". $str . '|';
	      } else {
	          $unique .= $name."=".' |';
	      }
	    
	      if ($$names[$i+1] !~ /^$n0/i) {
	          $unique .= "|";
		  $n0 = substr($$names[$i+1], 0, length($n0));
	      }
           }

           #--- Add display parameters image and data product check box
           my $cunique  = Browse::CGI::Encode($unique);
	   my $dlink=  $self->createSelectDPBox($unique,$cunique, $shortname);

           #--- Display a row of data 
           print "$dlink";
           $self->displayTextLine($line, $nskip, $longs,$delim);
        }
    	
        my $s = "";
        if ($$self{"rowcount"} > 1) {
           $s = "s";
        }
        print "</font></pre>\n";
        print "$$self{rowcount} row$s retrieved from $shortname<p>\n";

        #--- Link to Data Products options further on down the page
        print "&nbsp;&nbsp;<a href='#dataproducts'>Data Products Retrieval</a>";
     
        # Add plot button
        (my $stable = $table) =~ s/$::local_origin//g;
        print "&nbsp;&nbsp;<input type='submit' name='Plot $table' " .
	   "value='Plot' onclick=\"this.form.target='_blank'\"> your " .
	   "$stable results? <a target='helpwin' href='$::w3bhelp#Java_Plot'>" .
	   "(help)</a>" unless ($table =~ / x / and $table =~ /\//);
        # Add Cross Cor button
        print " &nbsp;&nbsp;<input type='submit' name='Correlate $table' " .
	   "value='Cross-correlate' onclick=\"this.form.target='_blank'\"> " .
	   "your results with another catalog or table? " .
	   "<a target='helpwin' href='$::w3bhelp#cross_query'>(help)</a><br>"
	    unless $table =~ / x / ; 


        $$self{rowcount} = 0;
    	
        my $dp   = $$meta{"dpsets"};
        if (scalar(keys %$dp)) {
           $self->{"dpavailable"} = 1;
        }
    }
}

sub displayTextLine {

    my ($self,$line, $skip, $longs, $delim) = @_;
    
    my $types=$$self{types};
    print "$delim";
    my $i;
    for ($i=$skip; $i <= $#$line; $i += 1) {
          
        if ($longs && grep($_ == $i, @$longs)) {
	    next;
	}
	my $colval=$$line[$i];
        #--- Encode < and > so value not interpreted as HTML tag
        if ($$types[$i] == 1) {$colval=HTML::Encode($colval);}
        #print $$line[$i]."$delim";
        print $colval."$delim";
    }
    print "\n";
    
    for ($i=0; $i <= $#$longs; $i += 1) {
        my @spl = split("\0", $$line[$$longs[$i]]);
	foreach (@spl) {
	    print "      $delim".$_."\n";
	}
	print "\n";  # Extra line to separate multiple long fields.
    }
}
#--- Display some options before displaying tables
sub selectAll {
   print "<table cellspacing='3' cellpadding='3'><tr><td valign='bottom'>";
   print "<script type=\"text/javascript\" language=\"JavaScript1.2\">" .
	"<!--\ndocument.write(\"<input type=\\\"checkbox\\\" " .
	"name=\\\"allfiles\\\" id=\\\"allfiles\\\" " .
	"onClick=\\\"var a=this.form.elements;" .
	"for (var i=0; i< a.length; i++){ " .
	"if (a[i].name.indexOf('Target') ==0) { " .
	"a[i].checked=this.checked }}\\\"> " .
	"<label for=\\\"allfiles\\\">Select all data products for all rows" .
	"</label>\")\n// -->\n</script>";
   print "</td><td>&nbsp;</td><td valign='bottom'>";
   print "<a href='#dataproducts'>Data Products Retrieval Options</a>";
   print "</td></td></table>";
}

# Print a somewhat reduced table header.

sub tableHeader {
    my ($self) = @_;
    
    my $textdisplay = $$self{"text"};
    my ($meta) = $$self{"meta"};
    my $results = $self->results();
    if ($results) {
       $results->DisplayLinks($meta);
    }

    #Browse::Results::DisplayLinks($meta);

    my $table_info = $$meta{"tableinfo"};
    my $table      = $$meta{"table"};
    
    $$self{"meta"} = $meta;

    my $radius;
    if ($$meta{"positions"}) {
	my $pos = $$meta{"positions"};
	my $pos1 = $$pos[0];
	my ($ra, $dec, $rad) = split('\|', $pos1);
	$radius = $rad;
    }

    $$self{"rowcount"} = 0;
    if ($$meta{"resource_name"} && $$meta{"resource_name"} ne $$self{"last_resource"}) {
        my $url = $$meta{"query_url"};
	$url =~ s/asu-xml/VizieR/;
        print "<hr>VizieR resource: <a target=helpwin href='$url'>$$meta{resource_title} ($$meta{resource_name})</a>\n";
	$$self{"last_resource"} = $$meta{"resource_name"};
    }
    #Browse::Results::PrintTableName($table, $table_info, '', $textdisplay);
    Browse::Results::PrintTableName($table, $table_info, $radius, $textdisplay);
}
sub pageFooter {
    my ($self) = @_;
   
    $self->closeTable();
    print "<hr><p>";
    #-- Display data products options if there are data products associated
    #-- with any  of the displayed tables
    if ($self->{"dpavailable"} ) {
       displayDP();
    } else {
       displayNoDP();
    }

    Browse::Results::Footer();
    Browse::Decor::EmailFooter();
    Browse::Results::Heraform($$self{"input"});
    print "</body></html>\n";
}
#--- Display data products retrieval options
sub displayDP {
print<<EOT2;
<a name='dataproducts'></a>
<span class='large'><b>Data Products</b></span><br>

<br>
<table border=0>
       <tr class='othertable' ><td class='othertable'>
        <div class="small"><strong> To Download or View data click a button below:</strong></div><p>
</td></tr>
<tr><td align='right'>
<input type='submit' name="files" onClick="this.form.target='dpwindow';setcheckDP(1);return(pageCheck())" value="Create Download Script"></td><td> data products for selected rows</td></tr>
       <tr class='othertable' ><td class='othertable' align='right'><input type='submit' name='files' onClick="this.form.target='dpwindow';setcheckDP(1);return(pageCheck())" value='Preview and Retrieve'>
       <input type='hidden' name='CheckSize' value='on'>
       <input type='hidden' name='showgifs' value='on'>
       </td><td> data products for selected rows</td></tr>
       <tr class='othertable' ><td class='othertable' align='right'><input type='submit' name='Save to Hera' value='Save Data Products to Hera' onclick="this.form.target='dpwindow'"></td><td><a target='helpwin' href='http://heasarc.gsfc.nasa.gov/hera'>What is Hera?</a></td></tr>
       <tr class='othertable' ><td class='othertable' align='right'><input type='submit' name="files" onClick="this.form.target='dpwindow';setcheckDP(1);return(pageCheck())" value="Retrieve"></td>
<td> data products for selected rows</td>
</tr>
       <tr><td colspan='2'>
     <a target='helpwin' href='$::w3bhelp#data_query'>Data Products file name filter (optional)</a> <input type="text" name="dpfilefilter_" id="dpfilefilter" value='' onkeypress="return noenter(event)"> <br>
       </td>
       </tr>
        
EOT2
    print "    </table>\n";

}
# Overload this class method to display Redisplay options 
sub YourSearchWas {
   my ($self,$positions,$qualifiers,$sorts,$dirs,$resmsg,$tables) = @_;


   my $imageHTML .= "<table><tr><td align='center' valign='middle'>\n";

   my $searchDisplayer = Browse::SearchBasedOn->new();
   $searchDisplayer->CoordSys($self->{coordsys}, $self->{equinox});

   #--- only display user entered parameter qualifiers
   my @nonzz = grep !/^zz/, @$tables;
   if (scalar (@nonzz) > 0) {
      $searchDisplayer->Qualifiers($qualifiers);
   }

   $searchDisplayer->Positions($positions);
   $searchDisplayer->MaxRows($self->MaxRows());
   if ($resmsg =~ /\w/) {
      $searchDisplayer->Resmsg($resmsg);
   }
   if ($self->{savedquery} =~ /\w/) {
      $searchDisplayer->SavedQuery($self->{savedquery});
   }

   $imageHTML .= $searchDisplayer->displaySearch();
 
   #--- get the display mode to make sure it does not change upon re-query
   my $displayformat= ref($self);
   $displayformat=~ s/Browse\:\://;
   $imageHTML .= "<input type=hidden name='requerydisplay' value='" .
	$displayformat ."'>\n";

   
   $imageHTML .= "</td><td>\n";
   print "$imageHTML";

   my $string = "<table>";
   $string .= $self->displayFormatOptions();

   $string .="<tr><td align='left' colspan='2'><input type='submit' " .
        "name='redisplay' value='Save All Objects To File' " .
        "onClick=\"noNewWindowSubmit(this.form)\"></td></tr>\n";


   $string .="<tr><td align='left' colspan='2'><input type='reset'></td></tr>";
   $string .= "</table></td></tr></table><p>";
   print $string;
}

#--- Message display when there are no data products
sub displayNoDP {
print<<EOT2;
<a name='dataproducts'></a>
<span class='large'><b>Data Products</b></span><br>
There are no data products available for selected tables.
EOT2
}

#--- overwrite class routine to add text below summary and before results
sub resultsSummary {
   my ($self, $tinfo, $tables) =@_;
   $self->SUPER::resultsSummary($tinfo, $tables);
   print "<p>";
   selectAll();
}

sub legend {}
sub Directions {}
sub beforeNewTable {}
sub afterTable {}

## Successful load
1;
