use strict;

package Browse::VODisplay;

# Display the results in VOTable  XML.

use vars qw( @ISA );
@ISA = qw( Browse::TextDisplay );
use Browse::TextDisplay;


# set/get index of search_offset column
sub offsetIndex {
   my ($self, $index) = @_;
   $$self{'offsetindex'} = $index if defined $index; 
   return $$self{'offsetindex'};
}

# Indicate that we're going to have a VOTable XML output.
sub pageHeader {
    my ($self) = @_;
    
    $self->setCoordFormat("decimal");
    $self->setDateFormat("mjd");
    
    $$self{table_count} = 0;
    
    print "Content-type: text/xml charset=\"utf-8\"\n" .
	"Content-Disposition: attachment; filename=votable.xml\n\n";
print <<EOT;
<?xml version="1.0"?>
<!DOCTYPE VOTABLE SYSTEM "http://us-vo.org/xml/VOTable.dtd">
<VOTABLE version="1.0">
EOT
		  
}

# Print the table header up till the parameter list.
# We defer that until we know the order of output parameters.
sub tableHeader {

    my ($self) = @_;
    
    $$self{tableopen} = 1;
    
    #print "self:<br>"  if $::debug;
    #Browse::Utils::DumpRef($self) if $::debug;
    if ($$self{tableCount} == 0) {
        my $system;
	my $equinox;
    
        # Handle the definitions -- these should only be printed once
	# for the entire request.
	
        if (lc($$self{coordsys}) eq "equatorial") {
	    if ($$self{equinox} <= 1975) {
	        $system="eq_FK4";
	    } else {
	        $system="eq_FK5";
	    }
	    $equinox = $$self{equinox}
	    
	} elsif (lc($$self{coordsys}) eq "galactic") {
	    $system="galactic";
	    
	}
	if ($system)  {
	
	    my $cstring = "system=\"$system\"";
	    if ($equinox) {
	        $cstring .= " equinox=\"$equinox\"";
	    }
	    
	    print <<EOT;
  <DEFINITIONS>
    <COOSYS $cstring />
  </DEFINITIONS>
EOT

        }
    }
    
    $$self{tableCount} += 1;
    
    my $resid = 9000+$$self{tableCount};
    my $meta  = $$self{"meta"};
    #print "meta:<br>"  if $::debug;
    #Browse::Utils::DumpRef($meta) if $::debug;
  
    my $tab   = $$meta{"tableinfo"}; 
    my $nam   = $$tab{"table_name"};
    my $desc  = $$tab{"table_description"};
    my $url   = $$tab{"table_document_url"};
    
    # Since Browse resources are exactly one table the names and titles
    # of the resource are redundant with the table.
    
    my $rad = $$tab{"default_search_radius"}/60;

    $desc =~ s/\>/\&gt;/g;
    $desc =~ s/\</\&lt;/g;
    $desc =~ s/\&/\&amp;/g;

    print <<EOT;
    
  <RESOURCE ID="T$resid">
     <DESCRIPTION>
       HEASARC Browse data service
       Please send inquiries to mailto:request\@athena.gsfc.nasa.gov
    </DESCRIPTION>
    <PARAM ID="default_search_radius" ucd="OBS_ANG-SIZE" value="$rad" ></PARAM>
    <TABLE>
      <DESCRIPTION> $desc </DESCRIPTION>
EOT

    $$self{rowCount} = 0;

}

sub legend {}

# Terminate the 'page' by closing the
# ASTRO element.
sub pageFooter {
    print "</VOTABLE>\n";
}

# Process each line of data.  When we process the
# first line we can finish writing out the data header.

sub processData {

    my ($self, $data) = @_;

    #--- omit the dynamically generated column names that 
    #--- are used for the table links.  These column names 
    #--- share a common prefix  
    my $dlPrefix = $self->depLinkPrefix();

    # Print the field descriptions -- can't do
    # this earlier since we don't have the order until now.
    my $names = $$data{"names"};
    if ($$self{rowCount} == 0) {
	#my $names = $$data{"names"};
	my $types = $$data{"types"};
        $self->printFields($names, $types, $dlPrefix);
    }
	
    # Print the data.
    
    $$self{rowCount} += 1;
    
    if (!$$self{"transform"}) {
        $$self{"transform"} = $self->setTransform($data);
    }
    
    my $trans = $$self{"transform"};
    $data = $$data{"data"};
    Browse::TableFormat::transformLine($data, $trans);

    grep s/\&/&amp\;/g, @$data;
    grep s/\>/&gt\;/g,  @$data;
    grep s/\</&lt\;/g,  @$data;
    grep s/\000/\n/g,   @$data;

    # remove offset source name to keep double type
    my $offindex = $self->offsetIndex();
    if ($offindex) {
       $$data[$offindex] =~ s/\s*\(.*\)//;
    }

    print "<TR>";
    for (my $i=0; $i <= $#$data; $i += 1) {
        #--- Skip dependency links
        if ($$names[$i] =~ /$dlPrefix/) {
	   next;
	}
        print "<TD>$$data[$i]</TD>";
    }
    print "</TR>\n"
    #print "<TR><TD>",join("</TD><TD>", @$data), "</TD></TR>\n";
    
}

# Print the fields to be displayed.
sub printFields {

    my ($self, $names, $types, $dlPrefix) = @_;
    
    my $paraminfo = $$self{"meta"}{"paraminfo"};
	
    # Loop over each column
    my $i;	
    for ($i=0; $i <= $#$names; $i += 1) {
	
	my $name = $$names[$i];
	my $type = $$types[$i];
	my $unit = $$paraminfo{$name}{"parameter_unit"};
	my $desc = $$paraminfo{$name}{"parameter_description"};

        #--- Skip dependency links
        #print "----------->$name <br>" if $::debug;
        if ($name =~ /$dlPrefix/) {next;}
        #print "---->$name <br>" if $::debug;
	$desc =~ s/\&/\&amp\;/g;
	$desc =~ s/</\&lt\;/g;
	$desc =~ s/>/\&gt\;/g;

	my $form;
	    
	$unit =~ s/^ *//;
	$unit =~ s/ *$//;
	if (length($unit) <= 0) {
	    $unit = undef;
	} else {
	    $unit = "unit=\"$unit\"";
	}
	    
	    
	if ($type <= 1  || $name eq "class") {
	    $form=  'datatype="char" arraysize="*"';
	} elsif ($type eq 6) {
	    $form=  'datatype="double"';
	} elsif ($type eq 7) {
	    $form = 'datatype="float"';
	} else {
	    $form = 'datatype="int"';
	}
        #print "n=$name   t-$type f=$form<br>" if $::debug;
	    
	my $ucd = $$paraminfo{$name}{"parameter_ucd"};
	if (!defined($ucd)) {
	    $ucd = "";
	} else {
	    $ucd = "ucd=\"$ucd\"";
	}
	
        if ($name eq "w3__unique_id") {
            $name = substr($name, 4);
	    $desc = "Integer key";
	    $form = 'datatype="int"';
	    
        } elsif ($name eq "ra") {
	    $form = 'datatype="double"';
		
	} elsif ($name eq "dec") {
	    $form = 'datatype="double"';
		
	} elsif ($name eq "_Search_Offset") {
	    $desc= "Offset of target/observation from query center";
	    $name= substr($name, 1);
            # save index for source name removal later 
            $self->offsetIndex($i);  
	    
	}
	    
	print <<EOT;
      
      <FIELD name="$name" $form $unit $ucd>
        <DESCRIPTION> $desc </DESCRIPTION>
      </FIELD>
EOT
    }

#--- If this is a positional search but there are no 
#--- query results add the Search Offset field for consistency.
#--- This is somewhat of a kludge.  The offset column name should
#--- probably be in the names array variable.
if (!$self->offsetIndex()) {
   my $fields = $self->{"meta"}{"fields"};
   if ($fields) {
      my @offsetcheck=grep /_Search_Offset/, @$fields;
      if (scalar(@offsetcheck) >0) {
	print <<EOT;

      <FIELD name="Search_Offset" datatype="double">
        <DESCRIPTION> Offset from the specified location </DESCRIPTION>
      </FIELD>
EOT

      }
   }
}
	
    print <<EOT;
	
	
      <DATA>
        <TABLEDATA>
EOT
}

sub showEmpties {

    my ($self, $empties, $errs) = @_;
    
    # Ignore the empty tables since they
    # should already be handled...
    
    my $err;
    foreach $err(@$errs) {
    
        $$err[0] =~ s/\>/\&gt;/g;
        $$err[0] =~ s/\</\&lt;/g;
        $$err[0] =~ s/\&/\&amp;/g;
        $$err[2] =~ s/\>/\&gt;/g;
        $$err[2] =~ s/\</\&lt;/g;
        $$err[2] =~ s/\&/\&amp;/g;

	print <<EOT;
  <RESOURCE>
    <DESCRIPTION> Error in querying resource $$err[2]: $$err[0] </DESCRIPTION>
  </RESOURCE>
EOT
    }
}

sub closeTable {

    my ($self) = @_;
    
    if (!defined($$self{meta})) {
        return;
    }
    
    my $meta = $$self{meta};
    print "meta:<br>"  if $::debug;
    Browse::Utils::DumpRef($meta) if $::debug;

    #--- omit the dynamically generated column names that 
    #--- are used for the table links.  These column names 
    #--- share a common prefix  
    my $dlPrefix = $self->depLinkPrefix();
    
    if (!$$self{tableopen}) {
        $self->tableHeader();
	# Need to fake the name and type fields to do print fields.
	my $paraminfo = $$meta{paraminfo};
	#my $names = $$paraminfo{"#default#"};
        
	my $names = $$meta{"fields"};
        #my @sortednames = sort @$names;
        #my @sortednames =  @$names;
	
	# Add in the unique key field...
        grep { s/.*(w3__unique_id)/$1/} @$names;
	my $types = ();
        my $index=0;
	foreach my $param (@$names) {
	    my $info = $$paraminfo{$param};
	    my $form = $$info{"parameter_format"};

            #print "VO::$param  $form<br>" if $::debug;

	    if ($form =~ /char/) {
	        push(@$types, 0);
	    } elsif ($form =~ /float4/) {
	        push(@$types, 7);
	    } elsif ($form =~ /float8/) {
	        push(@$types, 6);
	    } else {
	        push(@$types, 2);
	    }

            #--- Set a flag if Search Ofset is already in the fields list
	    if ($param eq "_Search_Offset") {
               # save index for source name removal later 
               $self->offsetIndex($index);  
               #Kludge until I can match type with query that returns results
               $$types[$index] = 6;  
            }
	    ++$index;
	}
	#$self->printFields($names, $types, $dlPrefix);
	$self->printFields($names, $types, $dlPrefix);
    }
    
    print <<EOT;
        </TABLEDATA>
      </DATA>
    </TABLE>
  </RESOURCE>
EOT
    delete($$self{meta});
    delete($$self{tableopen});
}

sub resultsSummary  {}
sub YourSearchWas   {}
sub beforeNewTable {}
sub afterTable {}

## Successful load
1;

# Modified 12/1/2005 by TAM to return header information even
# if no results are returned.
# Modified 5/9/2007 to change id="v1.0" to  version="1.0"
