#=============================================================================#
# Module Name:    Astro::CoordConv
# Module Author:  Edward J. Sabol
# Creation Date:  1995/09
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
# This module contains subroutines for converting coordinates between
# different formats.
#=============================================================================#
#
# $Id: CoordConv.pm,v 1.25 2004/03/18 22:02:31 sabol Exp $
# $Log: CoordConv.pm,v $
# Revision 1.25  2004/03/18 22:02:31  sabol
# Fixed ecliptic longitude range check in Ecl2Decimal().
#
# Revision 1.24  2003/02/03 21:02:44  sabol
# Updated comments style.
#
# Revision 1.23  2001/08/09 19:21:11  sabol
# Removed old implementations of RA_Decimal2HMS(), RA_Decimal2DMS(), and
#   Dec_Decimal2DMS() completely.
#
# Revision 1.22  2001/08/09 19:19:50  sabol
# Added support for user-specified coordinates accuracy to RA_Decimal2HMS(),
#   RA_Decimal2DMS(), and Dec_Decimal2DMS(). Added Equ2Sexagesimal(),
#   Gal2Sexagesimal(), and Ecl2Sexagesimal() for convenience wrappers which
#   also test the input ranges.
#
# Revision 1.21  2001/08/08 18:51:52  sabol
# When parsing coordinates, change values formatted using Fortran "D"
#   scientific notation to the more commonly accepted "E" scientific notation.
#   Sybase doesn't grok "D" notation, for example.
#
# Revision 1.20  2001/05/11 20:42:09  sabol
# Added parsing of decimal degrees in scientific notation to HMS2Decimal()
#   and DMS2Decimal().
#
# Revision 1.19  2001/05/04 14:33:54  sabol
# Changed allowed input range for R.A./LII/Elon in Equ2Decimal(),
#   Gal2Decimal(), and Ecl2Decimal() from [0,360) to [-360,360], but the values
#   are then normalized to [0,360).
#
# Revision 1.18  2001/05/03 18:01:25  sabol
# Tightened up the regular expressions for minutes and seconds in HMS2Decimal()
#   and DMS2Decimal(). Improved range checking in Equ2Decimal() and added range
#   checking to Gal2Decimal() and Ecl2Decimal().
#
# Revision 1.17  2000/08/31 15:07:42  sabol
# Added support for $Astro::Config::decimal_accuracy in HMS2Decimal() and
#   DMS2Decimal() for formatting values in decimal degrees.
#
# Revision 1.16  2000/08/28 16:26:52  sabol
# Allow three digits for the degrees in DMS2Decimal(). Added some range
#   checking to Equ2Decimal(). Tweaked some debug statements to be more
#   efficient.
#
# Revision 1.15  1999/11/10 18:35:00  sabol
# Due to a change in Perl 5.005, the result of an sprintf() call with a
#   floating point format is considered to be tainted; such values are
#   now detainted automatically.
#
# Revision 1.14  1999/08/11 11:44:37  sabol
# Now uses 'vars' compiler directive to specify package-wide variables.
#   Moved all 'use' directives after the package declaration.
#
# Revision 1.13  1999/08/11 10:55:12  sabol
# Now optionally exports all functions using Exporter.
#
# Revision 1.12  1998/06/16 13:07:49  sabol
# Now uses Astro::Config. Changed &::Clean calls to &$Astro::Config::cleaner.
#   Replaced all "<br>" strings with $Astro::Config::eol.
#
# Revision 1.11  1998/06/09 17:33:03  sabol
# Added RA_Decimal2DMS function. An additional digit of precision is now kept
#   in RA_Decimal2HMS for consistency with HDBfilter.
#
# Revision 1.10.1  1997/05/01 15:28:04  sabol
# Changed ConvCoord.pl into Astro::CoordConv.pm.
#
# Revision 1.10  1996/12/19 23:03:02  sabol
# Renamed Cel2Decimal function to Equ2Decimal. Code is slightly more
#   efficient in places. Added RA_Decimal2HMS and Dec_Decimal2DMS functions.
#   These Perl functions are based on the C functions ra_deg2hms and
#   dec_deg2dms, created by Song Yom.
#
# Revision 1.9  1996/10/25 21:22:39  sabol
# Now compliant with Perl 5's strict compiler directive. Also, a decimal value
#   that starts with a period instead of a number is now allowed. I.e., ".5"
#   is a valid value.
#
# Revision 1.8  1996/09/06 20:20:18  sabol
# More code cleaning. Simplified a few regular expressions.
#
# Revision 1.7  1996/08/26 18:35:05  sabol
# Cleaned up code for consistent style across all files.
#
# Revision 1.6  1996/08/21 22:40:02  sabol
# Yet even more code cleaning.
#
# Revision 1.5  1996/07/09 22:19:56  sabol
# Cleaned up the code considerably. Removed redundant Clean subroutine which
#   was copied to utils.pl long ago.
#
# Revision 1.4  1996/06/25 21:00:48  sabol
# Updated all source code to be Perl 5-compatiable.
#
# Revision 1.1  1995/12/19 19:38:05  kressel
# W3Browse Initial CVS
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name & address intact.
#=============================================================================#

package Astro::CoordConv;

use strict;
use Astro::Config;

require Exporter;
use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION $debug);
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(Equ2Decimal Gal2Decimal Ecl2Decimal HMS2Decimal DMS2Decimal
		RA_Decimal2HMS RA_Decimal2DMS Dec_Decimal2DMS
		Equ2Sexagesimal Gal2Sexagesimal Ecl2Sexagesimal);
$VERSION = '1.10';

$debug = 0;

#------------------------------------------------------------------------------
sub Equ2Decimal {

    my($coord1,$coord2) = @_;

    $coord1 = &HMS2Decimal($coord1);
    $coord2 = &DMS2Decimal($coord2);

    if ($coord1 > 360.0 || $coord1 < -360.0) {
	$coord1 = 'Error: Invalid input range for Equatorial right ascension.';
    } elsif ($coord1 < 0.0) {
	$coord1 += 360.0;
    } elsif ($coord1 == 360.0) {
	$coord1 = 0.0;
    }
    if ($coord2 > 90.0 || $coord2 < -90.0) {
	$coord2 = 'Error: Invalid input range for Equatorial declination.';
    }

    print 'Equ2Decimal:  ra: ',$coord1,$Astro::Config::eol,
          'Equ2Decimal: dec: ',$coord2,$Astro::Config::eol if $debug;

    return($coord1,$coord2);

} #eosub--Equ2Decimal
#------------------------------------------------------------------------------
sub Equ2Sexagesimal {

    my($coord1,$coord2,$accuracy) = @_;

    print 'Equ2Sexagesimal: accuracy: ',$accuracy,$Astro::Config::eol,
          'Equ2Sexagesimal:  ra: ',$coord1,$Astro::Config::eol,
          'Equ2Sexagesimal: dec: ',$coord2,$Astro::Config::eol if $debug;

    if ($coord1 > 360.0 || $coord1 < -360.0) {
	$coord1 = 'Error: Illegal decimal degree right ascension value. ' .
	          'Could not convert to sexagesimal format.';
    } elsif ($coord1 < 0.0) {
	$coord1 += 360.0;
    } elsif ($coord1 == 360.0) {
	$coord1 = 0.0;
    }
    if ($coord2 > 90.0 || $coord2 < -90.0) {
	$coord2 = 'Error: Illegal decimal degree declination value. ' .
	          'Could not convert to sexagesimal format.';
    }
    if (defined($accuracy) && ($accuracy !~ /^(\d+(\.\d*)?|\.\d+)$/)) {
	$coord1 = $coord2 = 'Error: Specified coordinates accuracy is not a valid non-negative number.';
    }

    unless ($coord1 =~ /^Error:/ || $coord2 =~ /^Error:/) {
	$coord1 = &RA_Decimal2HMS($coord1,$accuracy);
	$coord2 = &Dec_Decimal2DMS($coord2,$accuracy);
    }

    print 'Equ2Sexagesimal: output  ra: ',$coord1,$Astro::Config::eol,
          'Equ2Sexagesimal: output dec: ',$coord2,$Astro::Config::eol if $debug;

    return($coord1,$coord2);

} #eosub--Equ2Sexagesimal
#------------------------------------------------------------------------------
sub Gal2Decimal {

    my($coord1,$coord2) = @_;

    my($lii) = &DMS2Decimal($coord1);
    my($bii) = &DMS2Decimal($coord2);

    if ($lii > 360.0 || $lii < -360.0) {
	$lii = 'Error: Invalid input range for Galactic longitude.';
    } elsif ($lii < 0.0) {
	$lii += 360.0;
    } elsif ($lii == 360.0) {
	$lii = 0.0;
    }
    if ($bii > 90.0 || $bii < -90.0) {
	$bii = 'Error: Invalid input range for Galactic latitude.';
    }

    print 'Gal2Decimal: lii ',$lii,$Astro::Config::eol,
          'Gal2Decimal: bii ',$bii,$Astro::Config::eol if $debug;

    return($lii,$bii);

} #eosub--Gal2Decimal
#------------------------------------------------------------------------------
sub Gal2Sexagesimal {

    my($coord1,$coord2,$accuracy) = @_;

    print 'Gal/Ecl2Sexagesimal: accuracy: ',$accuracy,$Astro::Config::eol,
          'Gal/Ecl2Sexagesimal: lon: ',$coord1,$Astro::Config::eol,
          'Gal/Ecl2Sexagesimal: lat: ',$coord2,$Astro::Config::eol if $debug;

    if ($coord1 > 360.0 || $coord1 < -360.0) {
	$coord1 = 'Error: Illegal decimal degree longitude value. ' .
	          'Could not convert to sexagesimal format.';
    } elsif ($coord1 < 0.0) {
	$coord1 += 360.0;
    } elsif ($coord1 == 360.0) {
	$coord1 = 0.0;
    }
    if ($coord2 > 90.0 || $coord2 < -90.0) {
	$coord2 = 'Error: Illegal decimal degree latitude value. ' .
	          'Could not convert to sexagesimal format.';
    }
    if (defined($accuracy) && ($accuracy !~ /^(\d+(\.\d*)?|\.\d+)$/)) {
	$coord1 = $coord2 = 'Error: Specified coordinates accuracy is not a valid non-negative number.';
    }

    unless ($coord1 =~ /^Error:/ || $coord2 =~ /^Error:/) {
	$coord1 = &RA_Decimal2DMS($coord1,$accuracy);
	$coord2 = &Dec_Decimal2DMS($coord2,$accuracy);
    }

    print 'Gal/Ecl2Sexagesimal: output lon: ',$coord1,$Astro::Config::eol,
          'Gal/Ecl2Sexagesimal: output lat: ',$coord2,$Astro::Config::eol
	    if $debug;

    return($coord1,$coord2);

} #eosub--Gal2Sexagesimal
#------------------------------------------------------------------------------
sub Ecl2Decimal {

    my($coord1,$coord2) = @_;

    my($elon) = &DMS2Decimal($coord1);
    my($elat) = &DMS2Decimal($coord2);

    if ($elon > 360.0 || $elon < -360.0) {
	$elon = 'Error: Invalid range for Ecliptic longitude.';
    } elsif ($elon < 0.0) {
	$elon += 360.0;
    } elsif ($elon == 360.0) {
	$elon = 0.0;
    }
    if ($elat > 90.0 || $elat < -90.0) {
	$elat = 'Error: Invalid range for Ecliptic latitude.';
    }

    print 'Ecl2Decimal: elon ',$elon,$Astro::Config::eol,
          'Ecl2Decimal: elat ',$elat,$Astro::Config::eol if $debug;

    return($elon,$elat);

} #eosub--Ecl2Decimal
#------------------------------------------------------------------------------
sub Ecl2Sexagesimal {

    return(&Gal2Sexagesimal(@_));

} #eosub--Ecl2Sexagesimal
#------------------------------------------------------------------------------
sub HMS2Decimal {

    my($input) = shift(@_);
    my($value);

    print 'HMS2Decimal: begin',$Astro::Config::eol if $debug;

    my($sign)         = '[+-]?';
    my($integer)      = $sign.'\d{1,3}';
    my($decimal_exp)  = '(\d\.\d*|\.\d+)[eEdD][+-]?\d{1,3}';
    my($decimal)      = $sign.'((\d{1,3}(\.\d*)?|\.\d+)|'.$decimal_exp.')';
    my($hr_min)       = $sign.'\d{1,2}\s+([0-5]?[0-9](\.\d*)?|\.\d+)';
    my($hr_min_sec)   = $sign.'\d{1,2}\s+([0-5]?[0-9])\s+([0-5]?[0-9](\.\d*)?|\.\d+)';

    local($_) = &$Astro::Config::cleaner($input);
    print 'HMS2Decimal: input: ',$_,$Astro::Config::eol if $debug;

    if (/^($integer)$/) { # coordinate is in integral hours
 	print 'HMS2Decimal: integer ',$_,$Astro::Config::eol if $debug;
 	$value = $_ * 15; # return decimal degrees
    } elsif (/^($decimal)$/) { # coordinate is decimal, so assume degrees
	s/[dD]/E/g; # change Fortran "D" notation to "E" notation
	print 'HMS2Decimal: decimal ',$_,$Astro::Config::eol if $debug;
	$value = $_; # return decimal degrees
    } elsif (/^($hr_min_sec)$/) {
	print 'HMS2Decimal: hrs_min_sec ',$_,$Astro::Config::eol if $debug;
	my($hrs,$min,$sec) = split(/\s+/,$_,3);
	if ($hrs =~ /^-/) { $min *= -1; $sec *= -1; }
	$value = ($hrs + ($min/60.0) + ($sec/3600.0)) * 15.0;
    } elsif (/^($hr_min)$/) {
	print 'HMS2Decimal: hr_min ',$_,$Astro::Config::eol if $debug;
	my($hrs,$min) = split(/\s+/,$_,2);
	if ($hrs =~ /^-/) { $min *= -1; }
	$value = ($hrs + ($min/60.0)) * 15.0;
    } else {
	$value = 'Error: Invalid sexagesimal (hh mm ss) or decimal format for coordinate input.';
	print $value,$Astro::Config::eol if $debug;
    }

    if ($value !~ /error/i) {
	$value = &$Astro::Config::cleaner(sprintf('%.' . int($Astro::Config::decimal_accuracy) . 'f',$value));
	($value =~ /^([-0-9\.\+ ]+)$/) && ($value = $1);	# detaint
    }
    print 'HMS2Decimal: value ',$value,$Astro::Config::eol if $debug;

    return($value);

} #eosub--HMS2Decimal
#------------------------------------------------------------------------------
sub DMS2Decimal {

    my($input) = shift(@_);
    my($value);

    print 'DMS2Decimal: begin',$Astro::Config::eol if $debug;

    my($sign)         = '[+-]?';
    my($decimal_exp)  = '(\d\.\d*|\.\d+)[eEdD][+-]?\d{1,3}';
    my($decimal)      = $sign.'((\d{1,3}(\.\d*)?|\.\d+)|'.$decimal_exp.')';
    my($deg_min)      = $sign.'\d{1,3}\s+([0-5]?[0-9](\.\d*)?|\.\d+)';
    my($deg_min_sec)  = $sign.'\d{1,3}\s+([0-5]?[0-9])\s+([0-5]?[0-9](\.\d*)?|\.\d+)';

    local($_) = &$Astro::Config::cleaner($input);

    if (/^($decimal)$/) { # coordinate is already in decimal degrees
	s/[dD]/E/g; # change Fortran "D" notation to "E" notation
	print 'DMS2Decimal: decimal ',$_,$Astro::Config::eol if $debug;
	$value = $_; # return decimal degrees
    } elsif (/^($deg_min_sec)$/) {
	print 'DMS2Decimal: deg_min_sec ',$_,$Astro::Config::eol if $debug;
	my($deg,$min,$sec) = split(/\s+/,$_,3);
	if ($deg =~ /^-/) { $min *= -1.0; $sec *= -1.0; }
	$value = $deg + ($min/60.0) + ($sec/3600.0);
    } elsif (/^($deg_min)$/) {
	print 'DMS2Decimal: deg_min ',$_,$Astro::Config::eol if $debug;
	my($deg,$min) = split(/\s+/,$_,2);
	if ($deg =~ /^-/) { $min *= -1; }
	$value = $deg + ($min/60.0);
    } else {
	$value = 'Error: Invalid sexagesimal (ddd mm ss) or decimal format for coordinate input.';
	print $value,$Astro::Config::eol if $debug;
    }

    if ($value !~ /error/i) {
	$value = &$Astro::Config::cleaner(sprintf('%.' . int($Astro::Config::decimal_accuracy) . 'f',$value));
	($value =~ /^([-0-9\.\+ ]+)$/) && ($value = $1);	# detaint
    }
    print 'DMS2Decimal: value ',$value,$Astro::Config::eol if $debug;

    return($value);

} #eosub--DMS2Decimal
#------------------------------------------------------------------------------
sub RA_Decimal2HMS {

    my($degree,$accuracy) = @_;

    $accuracy = $Astro::Config::decimal_accuracy unless (defined($accuracy));
    print 'RA_Decimal2HMS: begin',$Astro::Config::eol,
          'RA_Decimal2HMS: accuracy: ',$accuracy,$Astro::Config::eol if $debug;
    my($min_secs,$max_secs,$fmt);
    my($min_mins,$max_mins) = (0,0);
    if ($accuracy >= 7) {
	$min_secs = 0.00005;
	$max_secs = 59.99995;
	$fmt = "%02d %02d %07.4f";
    } elsif ($accuracy >= 6) {
	$min_secs = 0.0005;
	$max_secs = 59.9995;
	$fmt = "%02d %02d %06.3f";
    } elsif ($accuracy >= 5) {
	$min_secs = 0.005;
	$max_secs = 59.995;
	$fmt = "%02d %02d %05.2f";
    } elsif ($accuracy >= 4) {
	$min_secs = 0.05;
	$max_secs = 59.95;
	$fmt = "%02d %02d %04.1f";
    } elsif ($accuracy >= 3) {
	$min_secs = 0.5;
	$max_secs = 59.5;
	$fmt = "%02d %02d %02.0f";
    } elsif ($accuracy >= 2) {
	$min_mins = 0.05;
	$max_mins = 59.95;
	$min_secs = $max_secs = 0;
	$fmt = "%02d %04.1f";
    } elsif ($accuracy >= 1) {
	$min_mins = 0.5;
	$max_mins = 59.5;
	$min_secs = $max_secs = 0;
	$fmt = "%02d %02.0f";
    } else {
	$min_secs = $max_secs = 0;
	$fmt = "%04.1f";
    }
    my($value);
    if ($degree >= 0.0 && $degree < 360.0) {
	my($t) = $degree / 15.0;
	if ($accuracy < 1) {
	    my($decimal_hours) = $t;
	    print 'RA_Decimal2HMS: (before rounding) ',$decimal_hours,$Astro::Config::eol if $debug;
	    $value = sprintf($fmt,$decimal_hours);
	} else {
	    my($hours) = int($t);
	    $t = ($t - $hours) * 60.0;
	    if ($accuracy < 3) {
		my($decimal_mins) = $t;
		print 'RA_Decimal2HMS: (before rounding) ',$hours,' ',$decimal_mins,$Astro::Config::eol if $debug;
		if ($decimal_mins < $min_mins) { $decimal_mins = 0.0; }
		if ($decimal_mins >= $max_mins) {
		    $decimal_mins = 0.0;
		    $hours++;
		}
		if ($hours == 24) {
		    $hours = 0;
		}
		$value = sprintf($fmt,$hours,$decimal_mins);
	    } else {
		my($mins) = int($t);
		my($secs) = ($t - $mins) * 60.0;
		print 'RA_Decimal2HMS: (before rounding) ',$hours,' ',$mins,' ',$secs,$Astro::Config::eol if $debug;
		if ($secs < $min_secs) { $secs = 0.0; }
		if ($secs >= $max_secs) {
		    $secs = 0.0;
		    $mins++;
		}
		if ($mins == 60) {
		    $mins = 0;
		    $hours++;
		}
		if ($hours == 24) {
		    $hours = 0;
		}
		$value = sprintf($fmt,$hours,$mins,$secs);
	    }
	}
	($value =~ /^([-0-9\.\+ ]+)$/) && ($value = $1);
    } else {
	$value = 'Error: Illegal decimal degree right ascension value. ' .
	         'Could not convert to sexagesimal format.';
    }
    print 'RA_Decimal2HMS: value: ',$value,$Astro::Config::eol if $debug;

    return($value);

} #eosub--RA_Decimal2HMS
#------------------------------------------------------------------------------
sub RA_Decimal2DMS {

    my($degree,$accuracy) = @_;

    $accuracy = $Astro::Config::decimal_accuracy unless (defined($accuracy));
    print 'RA_Decimal2DMS: begin',$Astro::Config::eol,
          'RA_Decimal2DMS: accuracy: ',$accuracy,$Astro::Config::eol if $debug;
    my($min_secs,$max_secs,$fmt);
    my($min_mins,$max_mins) = (0,0);
    if ($accuracy >= 7) {
	$min_secs = 0.0005;
	$max_secs = 59.9995;
	$fmt = "%3d %02d %06.3f";
    } elsif ($accuracy >= 6) {
	$min_secs = 0.005;
	$max_secs = 59.995;
	$fmt = "%3d %02d %05.2f";
    } elsif ($accuracy >= 5) {
	$min_secs = 0.05;
	$max_secs = 59.95;
	$fmt = "%3d %02d %04.1f";
    } elsif ($accuracy >= 4) {
	$min_secs = 0.5;
	$max_secs = 59.5;
	$fmt = "%3d %02d %02.0f";
    } elsif ($accuracy >= 3) {
	$min_mins = 0.05;
	$max_mins = 59.95;
	$min_secs = $max_secs = 0;
	$fmt = "%3d %04.1f";
    } elsif ($accuracy >= 2) {
	$min_mins = 0.5;
	$max_mins = 59.5;
	$min_secs = $max_secs = 0;
	$fmt = "%3d %02.0f";
    } elsif ($accuracy >= 1) {
	$min_secs = $max_secs = 0;
	$fmt = "%4.1f";
    } else {
	$min_secs = $max_secs = 0;
	$fmt = "%3.0f";
    }
    my($value);
    if ($degree >= 0.0 && $degree < 360.0) {
	if ($accuracy < 2) {
	    print 'RA_Decimal2DMS: (before rounding) ',$degree,$Astro::Config::eol if $debug;
	    $value = sprintf($fmt,$degree);
	} else {
	    my($t) = $degree;
	    my($dd) = int($t);
	    $t = ($t - $dd) * 60.0;
	    if ($accuracy < 4) {
		my($decimal_mins) = $t;
		print 'RA_Decimal2DMS: (before rounding) ',$dd,' ',$decimal_mins,$Astro::Config::eol if $debug;
		if ($decimal_mins < $min_mins) { $decimal_mins = 0.0; }
		if ($decimal_mins >= $max_mins) {
		    $decimal_mins = 0.0;
		    $dd++;
		}
		if ($dd == 360) {
		    $dd = 0;
		}
		$value = sprintf($fmt,$dd,$decimal_mins);
	    } else {
		my($mins) = int($t);
		my($secs) = ($t - $mins) * 60.0;
		print 'RA_Decimal2DMS: (before rounding) ',$dd,' ',$mins,' ',$secs,$Astro::Config::eol if $debug;
		if ($secs < $min_secs) { $secs = 0.0; }

		if ($secs >= $max_secs) {
		    $secs = 0.0;
		    $mins++;
		}
		if ($mins == 60) {
		    $mins = 0;
		    $dd++;
		}
		if ($dd == 360) {
		    $dd = 0;
		}
		$value = sprintf($fmt,$dd,$mins,$secs);
	    }
	}
	($value =~ /^([-0-9\.\+ ]+)$/) && ($value = $1);
    } else {
	$value = 'Error: Illegal decimal degree longitude value. ' .
	         'Could not convert to sexagesimal format.';
    }
    print 'RA_Decimal2DMS: value ',$value,$Astro::Config::eol if $debug;

    return($value);

} #eosub--RA_Decimal2DMS
#------------------------------------------------------------------------------
sub Dec_Decimal2DMS {

    my($degree,$accuracy) = @_;

    $accuracy = $Astro::Config::decimal_accuracy unless (defined($accuracy));
    print 'Dec_Decimal2DMS: begin',$Astro::Config::eol,
          'Dec_Decimal2DMS: accuracy: ',$accuracy,$Astro::Config::eol if $debug;
    my($min_secs,$max_secs,$fmt);
    my($min_mins,$max_mins) = (0,0);
    if ($accuracy >= 7) {
	$min_secs = 0.0905;
	$max_secs = 59.9995;
	$fmt = "%s%02d %02d %06.3f";
    } elsif ($accuracy >= 6) {
	$min_secs = 0.005;
	$max_secs = 59.995;
	$fmt = "%s%02d %02d %05.2f";
    } elsif ($accuracy >= 5) {
	$min_secs = 0.05;
	$max_secs = 59.95;
	$fmt = "%s%02d %02d %04.1f";
    } elsif ($accuracy >= 4) {
	$min_secs = 0.5;
	$max_secs = 59.5;
	$fmt = "%s%02d %02d %02.0f";
    } elsif ($accuracy >= 3) {
	$min_mins = 0.05;
	$max_mins = 59.95;
	$min_secs = $max_secs = 0;
	$fmt = "%s%02d %04.1f";
    } elsif ($accuracy >= 2) {
	$accuracy = 2;
	$min_mins = 0.5;
	$max_mins = 59.5;
	$min_secs = $max_secs = 0;
	$fmt = "%s%02d %02.0f";
    } elsif ($accuracy >= 1) {
	$min_secs = $max_secs = 0;
	$fmt = "%s%04.1f";
    } else {
	$min_secs = $max_secs = 0;
	$fmt = "%s%02.0f";
    }
    my($value);
    if ($degree >= -90.0 && $degree <= 90.0) {
	my($sign) = ($degree < 0) ? '-' : '+';
	my($t) = abs($degree);
	if ($accuracy < 2) {
	    print 'Dec_Decimal2DMS: (before rounding) ',$sign,$t,$Astro::Config::eol if $debug;
	    $value = sprintf($fmt,$sign,$t);
	} else {
	    my($dd) = int($t);
	    $t = ($t - $dd) * 60.0;
	    if ($accuracy < 4) {
		my($decimal_mins) = $t;
		print 'Dec_Decimal2DMS: (before rounding) ',$sign,$dd,' ',$decimal_mins,$Astro::Config::eol if $debug;
		if ($decimal_mins < $min_mins) { $decimal_mins = 0.0; }
		if ($decimal_mins >= $max_mins) {
		    $decimal_mins = 0.0;
		    $dd++;
		}
		$value = sprintf($fmt,$sign,$dd,$decimal_mins);
	    } else {
		my($mins) = int($t);
		my($secs) = ($t - $mins) * 60.0;
		print 'Dec_Decimal2DMS: (before rounding) ',$sign,$dd,' ',$mins,' ',$secs,$Astro::Config::eol if $debug;
		if ($secs < $min_secs) { $secs = 0.0; }
		if ($secs >= $max_secs) {
		    $secs = 0.0;
		    $mins++;
		}
		if ($mins == 60) {
		    $mins = 0;
		    $dd++;
		}
		$value = sprintf($fmt,$sign,$dd,$mins,$secs);
	    }
	}
	($value =~ /^([-0-9\.\+ ]+)$/) && ($value = $1);
    } else {
	$value = 'Error: Illegal decimal degree declination value. ' .
	         'Could not convert to sexagesimal.';
    }
    print 'Dec_Decimal2DMS: value: ',$value,$Astro::Config::eol if $debug;

    return($value);

} #eosub--Dec_Decimal2DMS
#------------------------------------------------------------------------------
1;
#########
#  EOF  #
#########
