#=============================================================================#
# Module Name:    Astro::Object::Resolver::SIMBAD
# Module Author:  Edward J. Sabol
# Creation Date:  1997/05
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Object-oriented interface to the SIMBAD name resolver. Object inherits
#   from base class Astro::Object::Resolver.
#=============================================================================#
#
# $Id: SIMBAD.pm,v 1.7 2003/02/04 23:06:07 sabol Exp $
# $Log: SIMBAD.pm,v $
# Revision 1.7  2003/02/04 23:06:07  sabol
# Now correctly handles the scenario where there are multiple matches but one
#   of the matches does not specify coordinates.
#
# Revision 1.6  2003/02/03 21:40:39  sabol
# Updated comments style. Added support for returning multiple matches and
#   information on alternative object names, object type, and bibliographic
#   reference codes.
#
# Revision 1.5  2002/10/25 23:04:00  sabol
# Detects when SIMBAD returns bogus coordinates in the form of tilde (~)
#   characters and returns a proper error message.
#
# Revision 1.4  2001/05/11 20:30:45  sabol
# Now does a fork() and exec() (via open('-|')) to pass the arguments to the
#   name resolver executable without using a shell. Some additional characters
#   that were not allowed in SIMBAD names previously are now allowed. Moved
#   wildcards error message from parent object to here.
#
# Revision 1.3  1999/08/11 11:53:41  sabol
# Now uses 'vars' compiler directive to specify package-wide variables.
#   Moved all 'use' directives after the package declaration. SIMBAD binary now
#   falls back to /dbase/lib/simbad/snr if /www/htdocs/cgi-bin/W3Browse/bin/snr
#   doesn't exist or isn't executable for whatever reason.
#
# Revision 1.2  1998/06/16 15:46:07  sabol
# Now uses Astro::Config. Replaced all references to Astro::Utils with
#   Astro::Config. Replaced all "<br>" and "\n" strings with
#   $Astro::Config::eol.
#
# Revision 1.1  1998/06/10 11:29:14  sabol
# Updated comments.
#
# Revision 1.0  1997/05/01 15:28:04  sabol
# Initial release.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Object::Resolver::SIMBAD;

use strict;
use FileHandle;
use Astro::Config;
use Astro::CoordSys;

use vars qw(@ISA $USER $PASSWORD $SNR);
@ISA = qw(Astro::Object::Resolver);

# Symbol table aliases to Astro::Config for backwards compatibility with
# older code. Please configure values in Astro::Config.
*Astro::Object::Resolver::SIMBAD::USER = \$Astro::Config::SIMBAD_user;
*Astro::Object::Resolver::SIMBAD::PASSWORD = \$Astro::Config::SIMBAD_password;
*Astro::Object::Resolver::SIMBAD::SNR = \$Astro::Config::SIMBAD_executable;

#------------------------------------------------------------------------------
sub initialize {
    # Method arguments: Same as Astro::Object::Resolver::new(), except no
    #   'type' option.

    my($self) = shift(@_);
    my(%args) = @_;

    $self->{'name'} = 'SIMBAD';
    my($cache_dir) = (($::cgi =~ /W3Browse/i) ? "$::root/$::cgi/dbm" :
		      $Astro::Object::Resolver::default_cache_directory);
    $self->{'cache'} = $cache_dir . '/' . $self->{'name'};
    $self->{'use_cache'} = (defined($args{'use_cache'})
			    ? $args{'use_cache'} : 1);
    $self->{'cache_read_only'} = (defined($args{'cache_read_only'})
				  ? $args{'cache_read_only'} : 0);
    $self->{'coordsys'} = Astro::CoordSys->new('system' => 'Equatorial',
					       'equinox' => 2000);
    return($self);

} #eosub--Astro::Object::Resolver::SIMBAD::initialize
#------------------------------------------------------------------------------
sub do_resolve {

    my($self,$name) = @_;

    return('Error: Sorry, wildcards in object names to be resolved ' .
	   'by the SIMBAD name resolver are not supported. ' .
	   'Please modify your input.') if ($name =~ /[\*\?]/);

    my($output) = 'Error: Invalid object name. Certain characters are not allowed.';
    my $matches = [];
    if ($name =~ /^([^;<>&\"\`\\\|]+)$/) {
	$name = $1;
	print "SIMBAD: name: ",$name,$Astro::Config::eol if $::debug;
	$output = '';
	my @all_output = ();
	my($fh) = new FileHandle;
	my($pid) = $fh->open('-|');
	return("Error: Could not fork process for executing SIMBAD name resolver.")
	  unless (defined($pid));
	if ($pid) {	# parent
	    @all_output = <$fh>;
	    $fh->close();
	    undef $fh;
	    chomp(@all_output);
	    $output = $all_output[0];
	} else {	# child
	    open(STDERR,">&STDOUT");        # redirect stderr to stdout
	    exec($SNR,$USER,$PASSWORD,$name)
	      || die "Error: Could not exec $SNR: $!\n";
	}
	print "SIMBAD: output: ",$output,$Astro::Config::eol if $::debug;
	if (length($output)) {
	    if ($output =~ /^\s+~\s+~\s+\(~\)/) {
		$output = 'error: coordinates not present';
	    } elsif ($output !~ /error/i) {
		$output =~ s/^\s+//;
		$output =~ s/\s+$//;
		my($RAd,$Decd) = split(/\s+/,$output);
		$RAd  = &$Astro::Config::cleaner($RAd);
		$Decd = &$Astro::Config::cleaner($Decd);
		$output = "$RAd, $Decd";
		print "SIMBAD: returning (coord1, coord2) = ($output)",
		      $Astro::Config::eol if $::debug;

		my %dup_check = ();
		while (@all_output) {
		    my $match = {};

		    # Parse coordinates.
		    my($coords) = shift(@all_output);
		    if ($coords =~ /^\s+~\s+~\s+\(~\)/) {
			$match->{'coordinates'} = $coords =
			  'error: coordinates not present';
		    } else {
			$coords =~ s/^\s+//;
			$coords =~ s/\s+$//;
			my($RAd,$Decd) = split(/\s+/,$coords);
			$RAd  = &$Astro::Config::cleaner($RAd);
			$Decd = &$Astro::Config::cleaner($Decd);
			$match->{'coordinates'} = [ $RAd, $Decd ];
		    }

		    push(@$matches,$match);

		    # Parse object name(s).
		    if (@all_output
			&& ($all_output[0] =~ /^\s*Name.*:\s+\S+/)) {
			my($names) = shift(@all_output);
			$names =~ s/^\s*Name.*:\s+//;
			$names =~ s/\s+$//;
			$dup_check{$names,$coords}++;
			pop(@$matches) if ($dup_check{$names,$coords} > 1);
			my(@all_names) = split(/,\s+/,$names);
			$match->{'name'} = shift(@all_names);
			if (scalar(@all_names)) {
			    $match->{'alt_names'} = [ @all_names ];
			}
		    }
		    
		    # Parse object type/classification.
		    if (@all_output
			&& ($all_output[0] =~ /^\s*Type:\s+\S+/)) {
			my($type) = shift(@all_output);
			$type =~ s/^\s*Type:\s+//;
			$type =~ s/\s+$//;
			$match->{'type'} = $type;
		    }

		    # Parse reference bibcode.
		    if (@all_output
			&& ($all_output[0] =~ /^\s*Bibcode:\s+\S+/)) {
			my($bibcode) = shift(@all_output);
			$bibcode =~ s/^\s*Bibcode:\s+//;
			$bibcode =~ s/\s+$//;
			$match->{'bibcode'} = $bibcode;
		    }

		    # Skip until the next pair of coordinates.
		    while (@all_output
			   && ($all_output[0] =~ /^\s*[A-Za-z]\S*:\s+\S+/)) {
			shift(@all_output);
		    }
		}
		my $match;
		my @objects = ();
		foreach $match (@$matches) {
		    next unless (ref($match->{'coordinates'}));
		    my %constructor_args = ('coordinates' =>
					      $match->{'coordinates'},
					    'coordsys' =>
						       $self->{'coordsys'});
		    if ($match->{'name'}) {
			$constructor_args{'name'} = $match->{'name'};
		    }
		    my $object = Astro::Object->new(%constructor_args);
		    if ($match->{'type'}) {
			$object->set_attribute('type' => $match->{'type'});
		    }
		    if ($match->{'bibcode'}) {
			$object->set_attribute('bibcode' =>
					         $match->{'bibcode'});
		    }
		    if ($match->{'alt_names'}) {
			$object->set_attribute('alt_names' =>
					         $match->{'alt_names'});
		    }
		    push(@objects,$object);
		}
		$matches = [ @objects ];
	    }
	} else {
	    return("Error: A problem was encountered executing the " .
		   "SIMBAD name resolver client.");
	}

	return("Error: SIMBAD name resolver temporarily unreachable. " .
	       "Please try again later or try a different name resolver.")
	  if ($output =~ /error: could not connect/i);

	return("Error: Sorry, object name \"$name\" is not recognized by the " .
	       "SIMBAD name resolver. Please go back and modify your input " .
	       "or try a different name resolver.")
	  if ($output =~ /error.*(unknown|not present|not recognized|identifier is not correctly written)/i);
    }

    return($output,$matches);

} #eosub--Astro::Object::Resolver::SIMBAD::do_resolve
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
