#=============================================================================#
# Module Name:    Astro::Utils
# Module Author:  Edward J. Sabol
# Creation Date:  1997/05
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Miscellaneous astronomical and/or astrophysics routines used by
#   the Astro module.
#=============================================================================#
#
# $Id: Utils.pm,v 1.14 2004/03/18 22:04:14 sabol Exp $
# $Log: Utils.pm,v $
# Revision 1.14  2004/03/18 22:04:14  sabol
# Renamed display_match()'s 'url_generator' parameter to 'link_generator'.
#   Changed 'max_alt_names' default to 7. The link generator code reference
#   is now passed only the match object.
#
# Revision 1.13  2004/03/17 01:41:40  sabol
# Added support for 'max_alt_names' parameter to display_matches() and made a
#   few tweaks to the implementation.
#
# Revision 1.12  2004/03/16 04:16:27  sabol
# Added display_matches() utility routine for formatting a list of resolver
#   matches in a consistent and reusable fashion.
#
# Revision 1.11  2003/12/02 19:24:04  sabol
# Protect against taking the square root of a negative number in sphrect().
#
# Revision 1.10  2003/02/03 21:05:15  sabol
# Updated comments style.
#
# Revision 1.9  2001/05/11 20:36:29  sabol
# Changed angsep() implementation to use the Haversine Formula (atan2 version).
#   The resulting values are the same, and the old implementation appeared to
#   be just a variant of the Haversine Formula, but the Haversine Formula is
#   well regarded and widely known, so might as well use that.
#
# Revision 1.7  1999/08/11 11:47:41  sabol
# Now uses 'vars' compiler directive to specify package-wide variables.
#   Moved all 'use' directives after the package declaration.
#
# Revision 1.6  1999/08/10 18:41:51  sabol
# Added some useful constants as package variables. Added angsep() and
#   sphrect() routines.
#
# Revision 1.5  1998/11/25 16:33:36  sabol
# Coco is no longer invoked in convert() if the input coordinate system is the
#   same as the output coordinate system.
#
# Revision 1.4  1998/11/25 14:19:59  sabol
# The equinox arguments sent to Coco in convert() now always have either a "B"
#   or "J" prefix if the coordinate system is equatorial.
#
# Revision 1.3  1998/06/16 13:33:06  sabol
# Now uses Astro::Config. Moved all package-wide configuration previously in
#   Utils to Astro::Config. Replaced all "<br>" and "\n" strings with
#   $Astro::Config::eol.
#
# Revision 1.2  1998/06/09 18:37:55  sabol
# $Astro::Utils::error_handler now defaults to \&CGI::Error if it's defined.
#
# Revision 1.1  1997/05/05 19:06:28  sabol
# Astro::Utils::convert should now be compatible with Perl 5's taint mode.
#
# Revision 1.0  1997/05/01 15:28:04  sabol
# Initial release.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Utils;

use strict;
use Astro::Config;

require Exporter;
use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION $PI $DTOR $MTOR $RTOM $RTOD);
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(convert clean angsep sphrect $PI $DTOR $MTOR $RTOM $RTOD);
$VERSION = '1.00';

#------------------------------------------------------------------------------
# Useful constants:
#
# Pi
$PI = 3.1415926535897932384626433832795028841971693993751;
# Degrees to Radians Conversion Constant (Pi/180)
$DTOR = 0.017453292519943295769236907684886127134428718885417;
# Arcminutes to Radians Conversion Constant (Pi/(180*60))
$MTOR = $DTOR / 60;
# Radians to Arcminutes Conversion Constant ((180*60)/Pi)
$RTOM = 180 * 60 / $PI;
# Radians to Degrees Conversion Constant (180/Pi)
$RTOD = 180 / $PI;

#------------------------------------------------------------------------------
sub convert {
    # Function  input: pair of astronomical coordinates (in decimal degrees!),
    #                    input coordinate system, output coordinate system
    # Function output: new pair of astronomical coordinates after converting
    #                    input coordinates to output coordinate system
    #                    or string describing any error

    my($coord1,$coord2,$input_coord_sys,$output_coord_sys) = @_;

    if (!ref($input_coord_sys) && !ref($output_coord_sys)) {
	return("Error: Not a valid coordinate system object.");
    }

    local($_);
    my($Ie) = $input_coord_sys->{'equinox'};
    my($Ict) = $input_coord_sys->{'coco'};
    my($Oe) = $output_coord_sys->{'equinox'};
    my($Oct) = $output_coord_sys->{'coco'};
    my($debug) = 0;

    return("Error: Input coordinates ($coord1,$coord2) are non-decimal.")
      if (($coord1 !~ /^[-0-9\.\+]+$/) || ($coord1 !~ /^[-0-9\.\+]+$/));

    ($coord1 =~ /^([-0-9\.\+]+)$/) && ($coord1 = $1);
    ($coord2 =~ /^([-0-9\.\+]+)$/) && ($coord2 = $1);

    return($coord1,$coord2) if ($input_coord_sys->equals($output_coord_sys));

    ($Ie =~ /^([-0-9BJ\.\+]+)$/) && ($Ie = $1);
    ($Oe =~ /^([-0-9BJ\.\+]+)$/) && ($Oe = $1);
    ($Ict =~ /^([0-9A-Za-z]+)$/) && ($Ict = $1);
    ($Oct =~ /^([0-9A-Za-z]+)$/) && ($Oct = $1);

    if ((($Ict eq '4') || ($Ict eq '5')) && ($Ie !~ /^[BJ]/i)) {
	$Ie = (($Ict eq '4') ? 'B' : 'J') . $Ie;
    }
    if ((($Oct eq '4') || ($Oct eq '5')) && ($Oe !~ /^[BJ]/i)) {
	$Oe = (($Oct eq '4') ? 'B' : 'J') . $Oe;
    }

    if ($debug) {
	print $Astro::Config::eol,"Coco: ",
	      "Converting $coord1, $coord2 in $Ict $Ie ",
	      "to $Oct $Oe",$Astro::Config::eol;
	print "Coco: Sending the following to $Astro::Config::coco:",
	      $Astro::Config::eol;
	print "Coco: F D",$Astro::Config::eol;
	print "Coco: I $Ict $Ie",$Astro::Config::eol;
	print "Coco: O $Oct $Oe",$Astro::Config::eol;
	print "Coco: $coord1 $coord2",$Astro::Config::eol;
    }

    # Execute coco binary with appropriate arguments.
    my($coco_arguments) = "<<eot\n" . "F D\n" . "I $Ict $Ie\n" .
      "O $Oct $Oe\n" . "$coord1 $coord2\n" . "eot\n";
    my(@coco_results) = `$Astro::Config::coco $coco_arguments`;
    &$Astro::Config::error_handler("Astro::Utils::convert: Coco execution failed!" . $Astro::Config::eol) if ($?);

    # Parse results.
    foreach (@coco_results) {
	chop;
	if (/^ ->\s*/) {
	    s/^ ->\s*//;
	    my(@fields) = split(/\s+/);
	    $coord1 = shift(@fields);
	    $coord2 = shift(@fields);
	    print "Astro::Utils::convert: Returning $coord1, $coord2",
	          $Astro::Config::eol if $debug;

	    return($coord1,$coord2);
	}
    }

    # If execution reaches here, something has gone wrong.
    return('Error: Undetermined problem encountered in coordinate conversion.' .
	   $Astro::Config::eol . "Coco: Tried to convert ($coord1,$coord2) " .
	   "in $Ict $Ie to $Oct $Oe" . $Astro::Config::eol .
	   "Coco: Sent the following to " . $Astro::Config::coco . ':' .
	   $Astro::Config::eol . 'Coco: F D' . $Astro::Config::eol .
	   "Coco: I $Ict $Ie" . $Astro::Config::eol . "Coco: O $Oct $Oe" .
	   $Astro::Config::eol. "Coco: $coord1 $coord2" . $Astro::Config::eol .
	   'Coco results follow:' . $Astro::Config::eol .
	   join($Astro::Config::eol,@coco_results));

} #eosub--Astro::Utils::convert
#------------------------------------------------------------------------------
sub clean {
    
    local($_) = join(' ',@_);

    # Clean the input string.
    s/\n+/ /g;   # change newlines to spaces
    s/^\s*//;	 # trim leading spaces
    s/^\+//;	 # trim leading plus sign
    s/\s*$//;	 # trim trailing spaces
    s/\s{2,}/ /; # squeeze embedded spaces

#    print "clean: $_$Astro::Config::eol" if $::debug;

    return($_);

} #eosub--Astro::Utils::clean
#------------------------------------------------------------------------------
sub angsep {
    # Function  input: two pair of spherical coordinates in decimal degrees
    # Function output: angle between the two points on a sphere in arcminutes
    
    my($lon1,$lat1,$lon2,$lat2) = @_;

    # This is an implementation of the Haversine Formula (using atan2) for
    # calculating the angular separation between two points on a unit sphere.
    # See <http://www.census.gov/cgi-bin/geo/gisfaq?Q5.1> for more info.

    $lon1 *= $DTOR;
    $lat1 *= $DTOR;
    $lon2 *= $DTOR;
    $lat2 *= $DTOR;
    my($dlon) = $lon2 - $lon1;
    my($dlat) = $lat2 - $lat1;
    my($a) = (sin($dlat/2))**2 + cos($lat1) * cos($lat2) * (sin($dlon/2))**2;
    my($b) = 1 - $a;
    if ($b < 0.0) {
	$b = 0.0;
    } else {
	$b = sqrt($b);
    }

    return(2 * atan2(sqrt($a),$b) * $RTOM);

} #eosub--Astro::Utils::angsep
#------------------------------------------------------------------------------
sub sphrect {
    # Perl version of FORTRAN routine created by Tom McGlynn.

    # Function  input: two pair of spherical coordinates in decimal degrees
    # Function output: four values in degrees: range of first coordinate,
    #                                          range of second coordinate

    my($r,$l0,$b0) = @_;
    ($r,$l0,$b0) = ($r * $DTOR,$l0 * $DTOR,$b0 * $DTOR);
    my($lmin,$lmax,$bmin,$bmax);

    # Given a point (l0,b0) and a radius of a circle, r, find the
    # smallest "rectangle" on the sphere than encloses the circle.
    # All quantities are expressed in radians.  It is assumed that
    # r>0 and l0, b0 are in the ranges [0,2 pi], [-Pi/2, Pi/2] respectively.
    # Note: If lmin > lmax, the range returned should
    # be [lmin,2 Pi] + [0,lmax].
    #
    # First check if a pole is enclosed in the circle. If so,
    # look at the full range of longitudes.

    if ((abs($b0) + $r) > $PI / 2) {
        $lmin = 0;
        $lmax = 2 * $PI;
    } else {

	# The maximum range in longitude occurs at the latitude:
	#    sin(b) = sin(b0)/cos(r)
	# Putting this in the radius function gives:

	my($deltal) = sqrt(cos($r)**2 - sin($b0)**2) / cos($b0);
	my($b) = 1 - $deltal**2;
	if ($b < 0.0) {	# Protect square root.
	    $b = 0.0;
	} else {
	    $b = sqrt($b);
	}
        $deltal = atan2($b,$deltal); # == acos($deltal)
        $lmin = $l0 - $deltal;
        $lmax = $l0 + $deltal;
        if ($lmin < 0) {
            $lmin += 2 * $PI;
	}
        if ($lmax > 2 * $PI) {
            $lmax -= 2 * $PI;
	}
    }

    $bmin = ($b0 - $r > -1 * $PI / 2) ? $b0 - $r : -1 * $PI / 2;
    $bmax = ($b0 + $r > $PI / 2) ? $PI / 2 : $b0 + $r;

    return($lmin * $RTOD,$lmax * $RTOD,$bmin * $RTOD,$bmax * $RTOD);

} #eosub--Astro::Utils::sphrect
#------------------------------------------------------------------------------
sub display_matches {
    # Function  input: hash with the following key/values:
    #       matches => reference to array of Astro::Object object as returned
    #                  by the resolver_matches method
    #       mode    => either 'html' or 'plain' (overrides autodetection)
    #       link_generator         => a code reference for generating links
    #       html_table_border      => integer (default 1)
    #       html_table_cellpadding => integer (default 5)
    #       html_table_cellspacing => integer (default 0)
    #       max_alt_names          => integer (default 7)
    # Function output: string containing nicely formatted listing of resolver
    #                  matches

    my(%args) = @_;

    my $matches = $args{'matches'};
    my $link_generator = exists($args{'link_generator'})
      ? $args{'link_generator'} : sub { shift->name; };
    my $mode = exists($args{'mode'})
      ? lc($args{'mode'}) : ($Astro::Config::eol =~ /<.+>/ ? 'html' : 'text');
    my $border = exists($args{'html_table_border'})
      ? sprintf('border=%d',int($args{'html_table_border'})) : 'border';
    my $cellspacing = exists($args{'html_table_cellspacing'})
      ? sprintf('cellspacing=%d',int($args{'html_table_cellspacing'})) : '';
    my $cellpadding = exists($args{'html_table_cellpadding'})
      ? int($args{'html_table_cellpadding'}) : 5;
    my $max_alt_names = exists($args{'max_alt_names'})
      && int($args{'max_alt_names'}) > 0 ? int($args{'max_alt_names'}) - 1 : 6;

    return('') unless (@$matches);

    my @matches = @{ $matches };
    my $first = shift(@matches);
    @matches = sort { $a->attribute('offset_from_primary_match') <=> $b->attribute('offset_from_primary_match') } @matches;
    @matches = ($first,@matches);

    my $coordsys = $first->coordsys->name;
    if ($coordsys eq 'Equatorial') {
	$coordsys = $first->coordsys->equinox;
	if ($matches->[0]->coordsys->sub_type eq 'FK5') {
	    $coordsys = 'J' . $coordsys;
	} elsif ($matches->[0]->coordsys->sub_type eq 'FK4') {
	    $coordsys = 'B' . $coordsys;
	}
    }

    my $text = '';
    if ($mode eq 'html') {

	$text = qq{<table id="AstroMatches" class="AstroMatches" $border $cellspacing cellpadding=$cellpadding>
<tr class="AstroMatches"><th class="AstroMatches">Identifier</th><th class="AstroMatches">Resolver</th><th class="AstroMatches">Coordinates ($coordsys)</th><th class="AstroMatches"><a class="AstroMatches" href="/Tools/name_resolver.html#object_types">Type</a></th><th class="AstroMatches">Alternative Names</th><th class="AstroMatches">Bibliographic Reference</th><th class="AstroMatches">Offset (arcmin)</th></tr>
};
	foreach my $match (@matches) {
	    my $name = &$link_generator($match) || '&nbsp;';
	    my $bibcode = $match->attribute('bibcode');
	    my @alts = ref($match->attribute('alt_names')) eq 'ARRAY'
	      ? @{ $match->attribute('alt_names') } : ();
	    my $alt_names = scalar(@alts)
	      ? 'Alt:' . join(', ',@alts[0..($#alts < $max_alt_names
					     ? $#alts : $max_alt_names)])
	      : undef;
	    my @info = map {
		if (defined($_)) {
		    s/&/&amp;/g;
		    if (/^(\[|Alt:)/) {
			s/^Alt://;
			s/ /&nbsp;/g;
			s/,&nbsp;/, /g;
		    }
		    s/</&lt;/g;
		    s/>/&gt;/g;
		    s/\"/&quot;/g;
		} else {
		    $_ = '&nbsp;';
		}
		$_;
	    } ($match->resolver_used,
	       '[ ' . join(', ',@{ $match->coordinates('format' => 'sexagesimal') }) . ' ]',
	       $match->attribute('type'),
	       $alt_names,
	       $match->attribute('bibcode'),
	       sprintf('%.3f',$match->attribute('offset_from_primary_match')));

	    my $offset = pop(@info);
 	    $info[$#info] = "<a href=\"http://adsabs.harvard.edu/cgi-bin/bib_query?$bibcode\">$bibcode</a>" if $bibcode;
	    unshift(@info,$name);
	    $text .= '<tr class="AstroMatches"><td class="AstroMatches">' . join('</td><td class="AstroMatches">',@info);
	    $text .= "</td><td class=\"AstroMatches\" align=right>$offset</td></tr>\n";
	}
	$text .= "</table>\n";

    } else {	# plain text

	my $fmt = "%-25s%-9s%-30s%s\n";
	$text = sprintf($fmt,'Identifier','Resolver',"    Coordinates ($coordsys)","Offset(\')");
	$text .= '-' x (length($text) - 1) . "\n";

	$fmt = "%-25s%-9s%-30s%9.3f\n";
	foreach my $match (@matches) {

	    my @info = map {
		$_ = defined($_) ? $_ : '';
	    } ($match->name,
	       $match->resolver_used,
	       '[ ' . join(', ', @{ $match->coordinates('format' => 'sexagesimal') }) . ' ]',
	       $match->attribute('offset_from_primary_match'));

	    $text .= sprintf($fmt,@info);
	}
    }

    return($text);

} #eosub--Astro::Utils::display_matches
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
