#=============================================================================#
# Module Name:    Astro::Utils::SLAConvert
# Module Author:  Edward J. Sabol
# Creation Date:  1999/09
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Convert coordinates from any given coordinate system to another
#   coordinate system using SLA routines.
#=============================================================================#
#
# $Id: SLAConvert.pm,v 1.5 2003/02/03 20:58:53 sabol Exp $
# $Log: SLAConvert.pm,v $
# Revision 1.5  2003/02/03 20:58:53  sabol
# Updated comments style.
#
# Revision 1.4  2001/05/11 20:47:25  sabol
# Added support for an optional Ecliptic epoch when converting to or from
#   the Ecliptic coordinate system.
#
# Revision 1.3  2001/05/02 14:47:33  sabol
# If the input and output coordinate systems and equinoxes are the same, then
#   just return the input coordinates.
#
# Revision 1.2  1999/11/10 18:35:25  sabol
# Due to a change in Perl 5.005, the result of an sprintf() call with a
#   floating point format is considered to be tainted; such values are
#   now detainted automatically.
#
# Revision 1.1  1999/09/02 16:05:53  sabol
# Added to Astro repository and released.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Utils::SLAConvert;

use strict;
use Astro::Utils qw($DTOR);
use Astro::Config;
use SLA;

$Astro::Config::converter = \&convert_using_SLA;

#------------------------------------------------------------------------------
sub convert_using_SLA {
    # Function  input: pair of astronomical coordinates (in decimal degrees!),
    #                    input coordinate system, output coordinate system
    # Function output: new pair of astronomical coordinates after converting
    #                    input coordinates to output coordinate system
    #                    or string describing any error

    my($coord1,$coord2,$input_coord_sys,$output_coord_sys) = @_;

    if (!ref($input_coord_sys) && !ref($output_coord_sys)) {
	return("Error: Not a valid coordinate system object.");
    }

    local($_);
    my($Ie) = $input_coord_sys->{'equinox'};
    my($Ict) = $input_coord_sys->{'coco'};
    my($Oe) = $output_coord_sys->{'equinox'};
    my($Oct) = $output_coord_sys->{'coco'};
    my($debug) = 0;

    return("Error: Input coordinates ($coord1,$coord2) are non-decimal.")
      if (($coord1 !~ /^[-0-9\.\+]+$/) || ($coord1 !~ /^[-0-9\.\+]+$/));

    ($coord1 =~ /^([-0-9\.\+]+)$/) && ($coord1 = $1);
    ($coord2 =~ /^([-0-9\.\+]+)$/) && ($coord2 = $1);

    return($coord1,$coord2) if ($input_coord_sys->equals($output_coord_sys));

    ($Ie =~ /^([-0-9BJ\.\+]+)$/) && ($Ie = $1);
    ($Oe =~ /^([-0-9BJ\.\+]+)$/) && ($Oe = $1);
    ($Ict =~ /^([0-9A-Za-z]+)$/) && ($Ict = $1);
    ($Oct =~ /^([0-9A-Za-z]+)$/) && ($Oct = $1);

    if ((($Ict eq '4') || ($Ict eq '5')) && ($Ie =~ /^[BJ]/i)) {
	$Ie =~ s/^[BJ]//i;
    }
    if ((($Oct eq '4') || ($Oct eq '5')) && ($Oe =~ /^[BJ]/i)) {
	$Oe =~ s/^[BJ]//i;
    }

    my($input_epoch_mjd,$output_epoch_mjd);
    if ($Ict eq 'E') {
	$input_epoch_mjd = $input_coord_sys->{'epoch_mjd'};
	if ($input_epoch_mjd eq '') {
	    $input_epoch_mjd = undef;
	}
    }
    if ($Oct eq 'E') {
	$output_epoch_mjd = $output_coord_sys->{'epoch_mjd'};
	if ($output_epoch_mjd eq '') {
	    $output_epoch_mjd = undef;
	}
    }

    if ($debug) {
	print $Astro::Config::eol,"SLA: ",
	      "Converting $coord1, $coord2 in $Ict $Ie ",
              (defined($input_epoch_mjd) ? "($input_epoch_mjd) " : ''),
	      "to $Oct $Oe",(defined($output_epoch_mjd) ?
			     " ($output_epoch_mjd)" : ''),$Astro::Config::eol;
    }

    if (($Ict eq $Oct) && ($Ie == $Oe)
	&& ((!defined($input_epoch_mjd) && !defined($output_epoch_mjd))
	    || ($input_epoch_mjd == $output_epoch_mjd))) {
	return($coord1,$coord2);
    }

    $coord1 *= $DTOR;
    $coord2 *= $DTOR;

    if (($Ict eq '4') && ($Oct eq '4')) {
	($coord1,$coord2) = &slaSubet($coord1,$coord2,$Ie);
	($coord1,$coord2) = &slaPreces('FK4',$Ie,$Oe,$coord1,$coord2);
	($coord1,$coord2) = &slaAddet($coord1,$coord2,$Oe);
    } elsif (($Ict eq '5') && ($Oct eq '5')) {
	($coord1,$coord2) = &slaPreces('FK5',$Ie,$Oe,$coord1,$coord2);
    } elsif (($Ict eq '4') && ($Oct eq '5')) {
	if ($Ie != 1950) {
	    ($coord1,$coord2) = &slaSubet($coord1,$coord2,$Ie);
	    ($coord1,$coord2) = &slaPreces('FK4',$Ie,1950,$coord1,$coord2);
	    ($coord1,$coord2) = &slaAddet($coord1,$coord2,1950);
	}
	($coord1,$coord2) = &slaFk45z($coord1,$coord2);
	if ($Oe != 2000) {
	    ($coord1,$coord2) = &slaPreces('FK5',2000,$Oe,$coord1,$coord2);
	}
    } elsif (($Ict eq '5') && ($Oct eq '4')) {
	if ($Ie != 2000) {
	    ($coord1,$coord2) = &slaPreces('FK5',$Ie,2000,$coord1,$coord2);
	}
	($coord1,$coord2) = &slaFk54z($coord1,$coord2);
	if ($Oe != 1950) {
	    ($coord1,$coord2) = &slaSubet($coord1,$coord2,1950);
	    ($coord1,$coord2) = &slaPreces('FK4',1950,$Oe,$coord1,$coord2);
	    ($coord1,$coord2) = &slaAddet($coord1,$coord2,$Oe);
	}
    } elsif (($Ict eq '5') && ($Oct eq 'G')) {
	if ($Ie != 2000) {
	    ($coord1,$coord2) = &slaPreces('FK5',$Ie,2000,$coord1,$coord2);
	}
	($coord1,$coord2) = &slaEqgal($coord1,$coord2);
    } elsif (($Ict eq '4') && ($Oct eq 'G')) {
	if ($Ie != 1950) {
	    ($coord1,$coord2) = &slaSubet($coord1,$coord2,$Ie);
	    ($coord1,$coord2) = &slaPreces('FK4',$Ie,1950,$coord1,$coord2);
	    ($coord1,$coord2) = &slaAddet($coord1,$coord2,1950);
	}
	($coord1,$coord2) = &slaFk45z($coord1,$coord2);
	($coord1,$coord2) = &slaEqgal($coord1,$coord2);
    } elsif (($Ict eq 'G') && ($Oct eq '5')) {
	($coord1,$coord2) = &slaGaleq($coord1,$coord2);
	if ($Oe != 2000) {
	    ($coord1,$coord2) = &slaPreces('FK5',2000,$Oe,$coord1,$coord2);
	}
    } elsif (($Ict eq 'G') && ($Oct eq '4')) {
	($coord1,$coord2) = &slaGaleq($coord1,$coord2);
	($coord1,$coord2) = &slaFk54z($coord1,$coord2);
	if ($Oe != 1950) {
	    ($coord1,$coord2) = &slaSubet($coord1,$coord2,1950);
	    ($coord1,$coord2) = &slaPreces('FK4',1950,$Oe,$coord1,$coord2);
	    ($coord1,$coord2) = &slaAddet($coord1,$coord2,$Oe);
	}
    } elsif (($Ict eq '5') && ($Oct eq 'E')) {
	if ($Ie != 2000) {
	    ($coord1,$coord2) = &slaPreces('FK5',$Ie,2000,$coord1,$coord2);
	}
	($coord1,$coord2) = &slaEqecl($coord1,$coord2,$output_epoch_mjd);
    } elsif (($Ict eq '4') && ($Oct eq 'E')) {
	if ($Ie != 1950) {
	    ($coord1,$coord2) = &slaSubet($coord1,$coord2,$Ie);
	    ($coord1,$coord2) = &slaPreces('FK4',$Ie,1950,$coord1,$coord2);
	    ($coord1,$coord2) = &slaAddet($coord1,$coord2,1950);
	}
	($coord1,$coord2) = &slaFk45z($coord1,$coord2);
	($coord1,$coord2) = &slaEqecl($coord1,$coord2,$output_epoch_mjd);
    } elsif (($Ict eq 'E') && ($Oct eq '5')) {
	($coord1,$coord2) = &slaEcleq($coord1,$coord2,$input_epoch_mjd);
	if ($Oe != 2000) {
	    ($coord1,$coord2) = &slaPreces('FK5',2000,$Oe,$coord1,$coord2);
	}
    } elsif (($Ict eq 'E') && ($Oct eq '4')) {
	($coord1,$coord2) = &slaEcleq($coord1,$coord2,$input_epoch_mjd);
	($coord1,$coord2) = &slaFk54z($coord1,$coord2);
	if ($Oe != 1950) {
	    ($coord1,$coord2) = &slaSubet($coord1,$coord2,1950);
	    ($coord1,$coord2) = &slaPreces('FK4',1950,$Oe,$coord1,$coord2);
	    ($coord1,$coord2) = &slaAddet($coord1,$coord2,$Oe);
	}
    } elsif (($Ict eq 'E') && ($Oct eq 'G')) {
	($coord1,$coord2) = &slaEcleq($coord1,$coord2,$input_epoch_mjd);
	($coord1,$coord2) = &slaEqgal($coord1,$coord2);
    } elsif (($Ict eq 'G') && ($Oct eq 'E')) {
	($coord1,$coord2) = &slaGaleq($coord1,$coord2);
	($coord1,$coord2) = &slaEqecl($coord1,$coord2,$output_epoch_mjd);
    } elsif (($Ict eq 'E') && ($Oct eq 'E')) {
	($coord1,$coord2) = &slaEcleq($coord1,$coord2,$input_epoch_mjd);
	($coord1,$coord2) = &slaEqecl($coord1,$coord2,$output_epoch_mjd);
    } else {
	return('Error: Undetermined problem encountered in ' .
	       'coordinate conversion.' . $Astro::Config::eol .
	       "Tried to convert ($coord1,$coord2) " .
	       "in $Ict $Ie to $Oct $Oe" . $Astro::Config::eol);
    }

    $coord1 /= $DTOR;
    $coord2 /= $DTOR;
    if ($coord1 < 0.0) {
	$coord1 += 360.0;
    }
    $coord1 = sprintf("%.8f",$coord1);
    $coord2 = sprintf("%.8f",$coord2);
    ($coord1 =~ /^([-0-9\.\+ ]+)$/) && ($coord1 = $1);
    ($coord2 =~ /^([-0-9\.\+ ]+)$/) && ($coord2 = $1);

    print "SLA: results = ($coord1, $coord2)",$Astro::Config::eol,
          $Astro::Config::eol if $debug;

    return($coord1,$coord2);

} #eosub--Astro::Utils::convert_using_SLA
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
