###############################################################################
#
# MODULE: $Id: DateTime.pm,v 1.31 2001/10/29 22:28:31 sabol Exp $
#
#  AUTHOR: Edward J. Sabol <sabol@alderaan.gsfc.nasa.gov>
# CREATED: 1996/10/09
#
# DESCRIPTION: This package contains a number of useful routines for
#  converting dates between various formats (MJD, SHF, YMD, Y.DOY, others) and
#  other miscellaneous date/time utility functions.
#
# HISTORY:
# $Log: DateTime.pm,v $
# Revision 1.31  2001/10/29 22:28:31  sabol
# Added hms2long() function for formatting hours/minutes/seconds.
#
# Revision 1.30  2001/06/08 22:18:06  sabol
# Minor fixes to some comments.
#
# Revision 1.29  2001/06/08 21:58:32  sabol
# Added utc2tt_secs(), tai2utc_secs(), tai2tt_secs(). Also, you can specify
#   an epoch date for tt2utc_secs() now.
#
# Revision 1.28  2001/06/08 21:52:17  sabol
# mjd2iso() no long appends " UTC" to the ISO-formatted time. The output time
#   system of mjd2iso() is actually the same as the input time system for the
#   MJD value, which IAU has declared to be TT unless specified otherwise.
#   Minor changes to the comments.
#
# Revision 1.27  2001/05/11 21:45:33  sabol
# Imported into HEASARC CVS repository.
#
# Revision 1.26  2001/01/19 17:30:15  sabol
# Added tt2utc_secs() for converting some number of seconds in TT to UTC.
#
# Revision 1.25  2001/01/18 15:02:28  sabol
# Now correctly handles the rounding of the seconds to the nearest second in
#   mjd2iso() when the number of seconds rounds up to 60.
#
# Revision 1.24  2000/12/06 18:27:20  sabol
# Fixed the regular expresion in ampm2military() so that it would not
#   incorrectly match invalid AM/PM times where the hour is zero.
#
# Revision 1.23  2000/12/06 18:18:54  sabol
# Added mdytime2mjd() and dmytime2mjd().
#
# Revision 1.22  2000/10/16 17:33:03  sabol
# Changed all occurrences of "shift(@_)" to just "@_" which should improve
#   efficiency slightly.
#
# Revision 1.21  2000/10/16 16:47:37  sabol
# Slightly more efficient operation in yd2ymd().
#
# Revision 1.20  2000/08/31 15:21:46  sabol
# Tweaked the regexp parsing in mdy2mjd() so that the comma is optional and
#   to allow for more than one space between the month and the day.
#
# Revision 1.19  2000/05/31 11:00:00  sabol
# Added dmy2mjd() to handle European-style dates. Improved the regular
#   expression used to parse the date in mdy2mjd().
#
# Revision 1.18  1999/09/29 14:56:50  sabol
# Changed the valid range for the $doy argument in yd2ymd() to be < 366 instead
#   of <= 365. This allows fractional days on the last day of the year to be
#   properly converted.
#
# Revision 1.17  1999/08/25 11:07:41  sabol
# Added jd2ymd() function that uses an algorithm from Claus Tondering's
#   Calendar FAQ at <http://www.pip.dknet.dk/~c-t/calendar.html>.
#
# Revision 1.16  1999/02/11 17:59:51  sabol
# Added mjd2time() and time2mjd() functions for converting between MJD and the
#   Unix time_t format (defined by POSIX as the number of non-leap seconds
#   since 1970-01-01 00:00:00 UTC).
#
# Revision 1.15  1999/02/11 16:42:22  sabol
# Enhanced dow(). If no arguments are supplied, then dow() behaves as it always
#   has and returns the day of the week for the current date (as provided by
#   localtime). However, dow() can now take an optional hash as an argument
#   which allows you to find the day of the week for an arbitrary date in
#   either MJD or Gregorian year, month, day formats. Added mdy2mjd() function
#   which converts old-style dates (e.g., 04/05/70, Apr-05-1970, etc.) to MJD.
#   month_name2num() now returns an error if the month name is less than three
#   characters in length. It also now works if you modify $[ -- "but don't do
#   that." :-)
#
# Revision 1.14  1999/02/01 16:45:13  sabol
# Added info regarding the 1999-01-01 leap second.
#
# Revision 1.13  1999/02/01 15:28:28  sabol
# Fixed a bug in yd2ymd() that caused erroneous results when the day of the
#   year was the last day of a month and was a non-integer (i.e. real) value.
#
# Revision 1.12  1998/12/09 12:00:29  sabol
# Minor modification to yd2ymd() that makes it slightly more efficient.
#
# Revision 1.11  1998/07/31 14:14:58  sabol
# yd2shfd() didn't work for years prior to 1900 or after 2100, so the range of
#   acceptable inputs has been modified accordingly. Consequently, ymd2mjd() no
#   longer uses yd2shfd(). Instead, ymd2mjd() uses a completely different
#   algorithm that is more accurate, more direct, and faster. Added ckint(),
#   ckfloat(), and ckymd() functions for better input checking across the
#   board. Many functions have been modified to use one of those three
#   functions. mjd2xte() and xte2mjd() no longer include leap seconds when
#   converting between the two formats. Added xte2shf() and shf2xte().
#   mjd2ymdhms() is also now more efficient.
#
# Revision 1.10  1998/07/24 11:27:28  sabol
# Fixed bug in yd2shfd() that caused off-by-one errors for years that come
#   before leap years prior to 1980.
#
# Revision 1.9  1998/06/11 16:54:27  sabol
# Added a lot more error checking and input checking to the conversion
#   functions.
#
# Revision 1.8  1998/06/11 10:27:41  sabol
# Improved mjd2yd() guessing by using 365.2422 as a more accurate measure of
#   how many days are in a year. Also, the guessing algorithm used in mjd2yd()
#   doesn't work for years less than 1 A.D., so it now bails out with an error
#   message if the guess is ever less than one to prevent an infinite loop.
#
# Revision 1.7  1998/02/24 17:50:22  sabol
# Added ascdate(), datestamp(), and local2mjd(). Updated leap seconds lists
#   with 1997-07-01.
#
# Revision 1.6  1998/02/19 18:11:35  sabol
# Added nint() function and used it in mjd2iso().
#
# Revision 1.5  1997/04/17 22:05:55  sabol
# Minor bug fix to iso2mjd() that prevented it from parsing the year, month,
#   and day when they were separated by dashes instead of slashes.
#
# Revision 1.3  1997/04/02 22:16:24  sabol
# Converting from MJD to SHF now rounds to the nearest integer instead of
#   simply truncating the floating point value to an integer.
#
# Revision 1.2  1997/02/04 00:07:21  sabol
# Added timestamp() and doy() routines.
#
# Revision 1.1  1996/12/10 23:35:55  sabol
# Added to W3Browse repository.
#
###############################################################################

use strict;

package DateTime;

#------------------------------------------------------------------------------
sub asctim {

    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $year += 1900;
    return(sprintf("%d-%02d-%02d %02d:%02d:%02d",
		   $year,$mon,$mday,$hour,$min,$sec));
} #eosub--asctim
#------------------------------------------------------------------------------
sub ascdate {

    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $year += 1900;
    return(sprintf("%d-%02d-%02d",$year,$mon,$mday));

} #eosub--ascdate
#------------------------------------------------------------------------------
sub timestamp {

    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $year += 1900;
    return(sprintf("%d%02d%02d-%02d%02d%02d",
		   $year,$mon,$mday,$hour,$min,$sec));
} #eosub--timestamp
#------------------------------------------------------------------------------
sub datestamp {

    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $year += 1900;
    return(sprintf("%d%02d%02d",$year,$mon,$mday));

} #eosub--datestamp
#------------------------------------------------------------------------------
sub now2mjd {

    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time);
    $mon++;
    $year += 1900;

    return(&ymdhms2mjd($year,$mon,$mday,$hour,$min,$sec));

} #eosub--now2mjd
#------------------------------------------------------------------------------
sub local2mjd {

    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $year += 1900;

    return(&ymdhms2mjd($year,$mon,$mday,$hour,$min,$sec));

} #eosub--local2mjd
#------------------------------------------------------------------------------
sub dow {
    # Function returns the day of the week as an integer with the following
    #    range of return values: 0-6 with 0 == Sunday, 1 == Monday, etc.
    # Input: optional hash with either of the following key/value pairs:
    #        'mjd' => integer or floating point MJD
    #        'ymd' => ref to an array of integers: [ year, month, day]
    #        If no arguments, then use current date as provided by localtime.

    my($dow);
    if (scalar(@_)) {
	my(%args) = @_;
	if (defined($args{'mjd'}) && ckfloat($args{'mjd'})) {
	    my($mjd) = int($args{'mjd'});
	    my($year,$month,$day) = &mjd2ymd($mjd);
	    $args{'ymd'} = [ $year, $month, int($day) ];
	}
	if (defined($args{'ymd'})) {
	    my($year,$month,$day) = @{ $args{'ymd'} };
	    if (&ckymd($year,$month,$day)) {
		# Algorithm comes from the Calendar FAQ at
		# <http://www.pip.dknet.dk/~c-t/cal/calendar20.txt>.
		my($a) = int((14 - $month) / 12);
		my($y) = $year - $a;
		my($m) = $month + 12*$a - 2;
		$dow = ($day + $y + int($y / 4) - int($y / 100) + int($y / 400)
			+ int((31 * $m) / 12)) % 7;
	    } else {
		$dow = 'error';
	    }
	} else {
	    $dow = 'error';
	}
    } else {
	$dow = (localtime(time))[6];
    }
    return($dow);

} #eosub--dow
#------------------------------------------------------------------------------
sub doy {
    # Function returns the day of the year.
    # Valid range of return values is 1-366. There is no day zero.

    return(((localtime(time))[7]) + 1);

} #eosub--doy
#------------------------------------------------------------------------------
sub ampm2military {
    # Input time string should contain something with the form:
    #   "12:23:45(AM|PM|am|pm)"
    # Return value with have the time in the form: "21:23:45"

    my($time) = @_;

    if ($time =~ /(0?[1-9]|1[0-2]):[0-5][0-9]:[0-5][0-9]\s*(am|pm)$/i) {
	$time =~ s/12(:[0-5][0-9]:[0-5][0-9])/00$1/;
	if ($time =~ /(AM|am)$/) {
	    $time =~ s/\s*(AM|am)$//;
	} elsif ($time =~ /(PM|pm)$/) {
	    $time =~ s/\s*(PM|pm)$//;
	    $time =~ s/^([^:]+ )?(0?[0-9]|1[0-2])(:.+)$/$1 . ($2 + 12) . $3/e;
	} else {
	    $time = 'error';
	}
    } else {
	$time = 'error';
    }

    return($time);

} #eosub--ampm2military
#------------------------------------------------------------------------------
sub hms2fday {

    my($hr,$min,$sec)= @_;

    return((((($sec / 60) + $min) / 60) + $hr) / 24);

} #eosub--hms2fday
#------------------------------------------------------------------------------
sub fday2hms {

    my($fday) = @_;

    my($hr,$min,$sec);
     
    $fday *= 86400;
    $hr = int($fday / 3600);
    $fday -= 3600 * $hr;
    $min = int($fday / 60);
    
    $fday -= 60 * $min;
    $sec = $fday;
     
    if (abs($sec - 60) < 0.00001) {
        $sec = 0;
        $min++;
    }
     
    if ($min == 60) {
        $min = 0;
        $hr++;
    }

    return($hr,$min,$sec);

} #eosub--fday2hms
#------------------------------------------------------------------------------
sub hms2long {

    my($hr,$min,$sec) = @_;
    
    $sec = &nint($sec);
    return(int($hr) . ' hour' . ($hr == 1 ? '' : 's') . ', ' .
	   int($min) . ' minute' . ($min == 1 ? '' : 's') . ', ' .
	   $sec . ' second' . ($sec == 1 ? '' : 's'));

} #eosub--hms2long
#------------------------------------------------------------------------------
sub mjd2yd {

    my($mjd) = @_;

    return('error','error') unless (&ckfloat($mjd));

    # There's probably a better, more exact solution for this.
    # For now, just use the inverse transformation with guessing.
    # [Algorithm by Tom McGlynn.]
    my($guessy) = int(($mjd - 44239.0) / 365.2422) + 1980;
    return('error','error') if ($guessy < 1);
    my($diy) = 365 + &leapyear($guessy);
    my($gmjd) = &ymd2mjd($guessy,1,1);

    while (($gmjd > $mjd) || (($gmjd + $diy) <= $mjd)) {
        if ($gmjd > $mjd) {
	    $guessy--;
	} else {
	    $guessy++;
	}
	return('error','error') if ($guessy < 1);
	$diy = 365 + &leapyear($guessy);
	$gmjd = &ymd2mjd($guessy,1,1);
    }

    return($guessy,1 + $mjd - $gmjd);

} #eosub--mjd2yd
#------------------------------------------------------------------------------
sub ymd2mjd {

    my($y,$m,$d) = @_;

    return('error') unless (&ckymd($y,$m,$d));

    my($aa) = 367 * $y;
    my($bb) = - int((7 * ($y + int(($m + 9) / 12))) / 4);
    my($cc) = - int(3 * (int(($y + int(($m - 9) / 7)) / 100) + 1) / 4);
    my($dd) = int(275 * $m / 9);
    my($ee) = $d + 1721029 - 2400000 - 1;
    
    my($mjd) = $aa + $bb + $cc + $dd + $ee;

# Old algorithm:
#     my($shfd) = &ymd2shfd(@_);
#     my($mjd) = $shfd;
#
#     if ($mjd ne 'error') {
#         $mjd += 44239;
#     }

    return($mjd);

} #eosub--ymd2mjd
#------------------------------------------------------------------------------
sub mjd2ymd {

    my($mjd) = @_;

    my($year,$day) = &mjd2yd($mjd);

    return(&yd2ymd($year,$day));

} #eosub--mjd2ymd
#------------------------------------------------------------------------------
sub ymdhms2mjd {

    # Note: This does not take the time system into account. If the input time
    # is TT or UTC, then the output time system will be the same.

    my($year,$month,$day,$hr,$min,$sec) = @_;

    my($mjd) = &ymd2mjd($year,$month,$day);
    if ($mjd ne 'error') {
	$mjd += &hms2fday($hr,$min,$sec);
    }

    return($mjd);

} #eosub--ymdhms2mjd
#------------------------------------------------------------------------------
sub mjd2ymdhms {

    # Note: This does not take the time system into account. If the input MJD
    # value is TT or UTC, then the output time system will be the same.

    my($mjd) = @_;

    my($year,$month,$day) = &mjd2ymd($mjd);
    my($fday) = $day - int($day);
    my($hour,$min,$sec) = &fday2hms($fday);
    if ($hour == 24) {
	($year,$month,$day,$hour,$min,$sec) = &mjd2ymdhms(&nint($mjd));
    }

    return($year,$month,$day,$hour,$min,$sec);

} #eosub--mjd2ymdhms
#------------------------------------------------------------------------------
sub ymd2shfd {
    # SHFD == days since 1980-01-01

    my($year,$month,$day) = @_;
    my($doy);

    ($year,$doy) = &ymd2yd($year,$month,$day);

    return(&yd2shfd($year,$doy));

} #eosub--ymd2shfd
#------------------------------------------------------------------------------
sub ymd2yd {

    my($year,$month,$day) = @_;

    my($doy);
    my(@dom) = (0,31,28,31,30,31,30,31,31,30,31,30,31);
    my($leap) = &leapyear($year);

    $dom[2] += $leap;
    if ($year ne 'error' && $month ne 'error' and $day ne 'error'
        && $year > 0 && $month > 0 && $month < 13 && $day > 0
        && $day <= $dom[$month]) {
	my($loop);
        for ($loop = 1,$doy = $day; $loop < $month; $loop++) {
            $doy += $dom[$loop];
        }
    } else {
        $year = $doy = 'error';
    }

    return($year,$doy);

} #eosub--ymd2yd
#------------------------------------------------------------------------------
sub yd2ymd {

    my($year,$doy) = @_;

    my($month,$day,$loop);
    my(@dom) = (0,31,28,31,30,31,30,31,31,30,31,30,31);
    my($leap) = &leapyear($year);

    if (($year > 0) && ($doy > 0) && ($doy < (366 + $leap))) {
	$dom[2] += $leap;
        for ($loop = 1,$month = 0,$day = $doy; !$month && $loop <= 12;
	     $loop++) {
	    if (($day - $dom[$loop]) < 1) {
		$month = $loop;
	    } else {
		$day -= $dom[$loop];
	    }
        }
    } else {
        $year = $month = $day = 'error';
    }

    return($year,$month,$day);

} #eosub--yd2ymd
#------------------------------------------------------------------------------
sub leapyear {

    my($year) = @_;

    return((($year % 4 == 0) && ($year % 100 != 0)) || ($year % 400 == 0));

} #eosub--leapyear
#------------------------------------------------------------------------------
sub yd2shfd {

    my($year,$doy) = @_;

    return('error') unless (&ckint($year) && &ckfloat($doy));

    my($shfd);
    if (($year >= 1900) && ($year < 2100) && ($doy > 0)) {
        my($yearlen) = 365 + &leapyear($year);
	if ($doy <= $yearlen) {
	    if ($year >= 1980) {
		$shfd = (($year - 1980) * 365 + int(($year - 1977) / 4)
			 + $doy - 1);
	    } else {
		$shfd = $doy - $yearlen - 1;
		$shfd -= (1979 - $year) * 365 + int((1979 - $year) / 4);
	    }
	} else {
	    $shfd = 'error';
	}
    } else {
        $shfd = 'error';
    }

    return($shfd);

} #eosub--yd2shfd
#------------------------------------------------------------------------------
sub month_name2num {

    my($month) = @_;

    return('error') if (length($month) < 3);
    my($abbrev) = substr($month,$[,3);
    my($month_names) = 'JanFebMarAprMayJunJulAugSepOctNovDec';
    my($pos) = index($month_names,$abbrev);
    if ($pos >= $[) {
	$month = (($pos - $[) / 3) + 1;
    } else {
	$month = 'error';
    }

    return($month);

} #eosub--month_name2num
#------------------------------------------------------------------------------
sub mdy2mjd {

    # Handles the following date formats:
    #    Apr-05-70
    #    Apr 5 1970
    #    April 5, 1970
    #    04/05/70
    #    4/5/70
    # In all cases either a two- or four-digit year is allowed.
    # If a two-digit year is supplied, it assumes the 20th century if it's
    # between 50 and 99 (inclusive) and the 21st century if it's between 0
    # and 49 (inclusive).
    # All uppercase and all lowercase also allowed for the month names.
    # Note: This does not take the time system into account. If the input time
    # is TT or UTC, then the output time system will be the same.

    my($date) = @_;
    my($mjd);

    if ($date =~ /^\s*(([A-Za-z])([A-Za-z]{2})[A-Za-z]*|(0?[0-9]|1[0-2]))[-\/ ] *([0-2][0-9]|3[01]|0?[1-9])([-\/]|,? +)(\d{2,4})\s*$/) {
	my($month,$month_name,$day,$year) = ($4,uc($2) . lc($3),$5,$7);
	# Handle two-digit years as well.
	if ($year >= 50 && $year < 100) {
	    $year += 1900;
	} elsif ($year >= 0 && $year < 50) {
	    $year += 2000;
	}
	$month = &month_name2num($month_name) unless (&ckint($month));
	$mjd = &ymd2mjd($year,$month,$day);
    } else {
	$mjd = 'error';
    }

    return($mjd);

} #eosub--mdy2mjd
#------------------------------------------------------------------------------
sub dmy2mjd {

    # Handles the following date formats:
    #    05-Apr-1970
    #    05-Apr-70
    #    5 Apr 1970
    #    5 April 70
    #    05/04/1970
    #    05/04/70
    #    5/4/70
    # In all cases either a two- or four-digit year is allowed.
    # If a two-digit year is supplied, it assumes the 20th century if it's
    # between 50 and 99 (inclusive) and the 21st century if it's between 0
    # and 49 (inclusive).
    # All uppercase and all lowercase also allowed for the month names.

    my($date) = @_;
    my($mjd);

    if ($date =~ /^\s*([0-2][0-9]|3[01]|0?[1-9])([-\/]| +)(([A-Za-z])([A-Za-z]{2})[A-Za-z]*|(0?[1-9]|1[0-2]))([-\/]| +)(\d{2,4})\s*$/) {
	my($month,$month_name,$day,$year) = ($6,uc($4) . lc($5),$1,$8);
	# Handle two-digit years as well.
	if ($year >= 50 && $year < 100) {
	    $year += 1900;
	} elsif ($year >= 0 && $year < 50) {
	    $year += 2000;
	}
	$month = &month_name2num($month_name) unless (&ckint($month));
	$mjd = &ymd2mjd($year,$month,$day);
    } else {
	$mjd = 'error';
    }

    return($mjd);

} #eosub--dmy2mjd
#------------------------------------------------------------------------------
sub mdytime2mjd {

    # Handles the following date formats:
    #    Apr-05-70  4:28PM
    #    Apr 5 1970 11:28:12AM
    #    April 5, 1970  4:28am
    #    04/05/70 04:28pm
    #    04/05/1970  14:28:12
    #    4/5/70  4:28:12am
    # In all cases either a two- or four-digit year is allowed.
    # If a two-digit year is supplied, it assumes the 20th century if it's
    # between 50 and 99 (inclusive) and the 21st century if it's between 0
    # and 49 (inclusive).
    # All uppercase and all lowercase also allowed for the month names and the
    # AM/PM designation.
    # Seconds are optional in the time portion of the input string.
    # Note: This does not take the time system into account. If the input time
    # is TT or UTC, then the output time system will be the same.

    my($date) = @_;

    $date =~ s/ +(0?[0-9]|1[0-9]|2[0-3])(:[0-5][0-9])(:[0-5][0-9])? *(AM|am|PM|pm)? *$//;
    my($hh,$mm,$ss,$ampm) = ($1,$2,$3,$4);
    return('error') if ($hh eq '');
    $ss = ':00' unless ($ss);
    my($time) = $hh . $mm . $ss . uc($ampm);
    if ($ampm) {
	$time = &DateTime::ampm2military($time);
	return($time) if ($time eq 'error');
    }
    my($mjd) = &DateTime::mdy2mjd($date);
    return($mjd) if ($mjd eq 'error');
    $mjd += &DateTime::hms2fday(split(/:/,$time));

    return($mjd);

} #eosub--mdytime2mjd
#------------------------------------------------------------------------------
sub dmytime2mjd {

    # Handles the following date formats:
    #    05-Apr-1970  4:28PM
    #    05-Apr-70 11:28:12AM
    #    5 Apr 1970  4:28am
    #    5 April 70 04:28pm
    #    05/04/1970  14:28:12
    #    05/04/70  14:28:12
    #    5/4/70  14:28
    # In all cases either a two- or four-digit year is allowed.
    # If a two-digit year is supplied, it assumes the 20th century if it's
    # between 50 and 99 (inclusive) and the 21st century if it's between 0
    # and 49 (inclusive).
    # All uppercase and all lowercase also allowed for the month names and the
    # AM/PM designation.
    # Seconds are optional in the time portion of the input string.
    # Note: This does not take the time system into account. If the input time
    # is TT or UTC, then the output time system will be the same.

    my($date) = @_;

    $date =~ s/ +(0?[0-9]|1[0-9]|2[0-3])(:[0-5][0-9])(:[0-5][0-9])? *(AM|am|PM|pm)? *$//;
    my($hh,$mm,$ss,$ampm) = ($1,$2,$3,$4);
    return('error') if ($hh eq '');
    $ss = ':00' unless ($ss);
    my($time) = $hh . $mm . $ss . uc($ampm);
    if ($ampm) {
	$time = &DateTime::ampm2military($time);
	return($time) if ($time eq 'error');
    }
    my($mjd) = &DateTime::dmy2mjd($date);
    return($mjd) if ($mjd eq 'error');
    $mjd += &DateTime::hms2fday(split(/:/,$time));

    return($mjd);

} #eosub--dmytime2mjd
#------------------------------------------------------------------------------
sub iso2mjd {

    # Note: This does not take the time system into account. If the input time
    # is TT or UTC, then the output time system will be the same.

    my($isodate) = @_;
    my($mjd);

    if ($isodate =~ /^\s*\d{4}(-|\/)(0[0-9]|1[0-2])(-|\/)([0-2][0-9]|3[01])\s+([01][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9](.\d+)?\s*$/) {
	my($date,$time) = split(/\s+/,$isodate,2);
	my($year,$month,$day) = split(/[-\/]/,$date);
	my($hour,$min,$sec) = split(/:/,$time);
	$mjd = &ymdhms2mjd($year,$month,$day,$hour,$min,$sec);
    } elsif ($isodate =~ /^\s*(\d{4})(-|\/)(0[0-9]|1[0-2])(-|\/)([0-2][0-9]|3[01])\s*$/) {
	my($year,$month,$day) = ($1,$3,$5);
	$mjd = &ymd2mjd($year,$month,$day);
    } else {
	$mjd = 'error';
    }

    return($mjd);

} #eosub--iso2mjd
#------------------------------------------------------------------------------
sub mjd2iso {

    # Note: This does not take the time system into account. If the input MJD
    # value is TT or UTC, then the output time system will be the same.

    my($mjd) = @_;
    my($iso);

    return('error') unless (&ckfloat($mjd));

    if (int($mjd) == $mjd) {
	my($year,$month,$day) = &mjd2ymd($mjd);

	if ($year eq 'error' || $month eq 'error' || $day eq 'error') {
	    $iso = 'error';
	} else {
	    $iso = sprintf("%04d-%02d-%02d",$year,$month,$day);
	}
    } else {
	my($year,$month,$day,$hour,$min,$sec) = &mjd2ymdhms($mjd);

	if ($year eq 'error' || $month eq 'error' || $day eq 'error') {
	    $iso = 'error';
	} elsif (abs($sec - &nint($sec)) < 0.05) {
	    $sec = &nint($sec);
	    if ($sec == 60) {
		$sec = 0;
		$min++;
	    }
	    if ($min == 60) {
		$min = 0;
		$hour++;
	    }
	    if ($hour == 24) {
		($year,$month,$day,$hour,$min,$sec) = &mjd2ymdhms(&nint($mjd));
	    }
	    $iso = sprintf("%04d-%02d-%02d %02d:%02d:%02d",
			   $year,$month,$day,$hour,$min,$sec);
	} else {
	    $iso = sprintf("%04d-%02d-%02d %02d:%02d:%04.1f",
			   $year,$month,$day,$hour,$min,$sec);
	}
    }

    return($iso);

} #eosub--mjd2iso
#------------------------------------------------------------------------------
sub mjd2tjd {

    my($mjd) = @_;

    return(&ckfloat($mjd) ? $mjd - 40000.0 : 'error');

} #eosub--mjd2tjd
#------------------------------------------------------------------------------
sub tjd2mjd {

    my($tjd) = @_;

    return(&ckfloat($tjd) ? $tjd + 40000.0 : 'error');

} #eosub--tjd2mjd
#------------------------------------------------------------------------------
sub mjd2jd {

    my($mjd) = @_;

    return(&ckfloat($mjd) ? $mjd + 2400000.5 : 'error');

} #eosub--mjd2jd
#------------------------------------------------------------------------------
sub jd2mjd {

    my($jd) = @_;

    return(&ckfloat($jd) ? $jd - 2400000.5 : 'error');

} #eosub--jd2mjd
#------------------------------------------------------------------------------
sub jd2ymd {
    # Based on algorithm from Claus Tondering's Calendar FAQ
    # at <http://www.pip.dknet.dk/~c-t/calendar.html>.

    my($jd) = @_;

    return('error') unless (&ckfloat($jd));

    my($a) = &nint($jd) + 32045;
    my($b) = int((4 * ($a + 36524)) / 146097) - 1;
    my($c) = $a - int(($b * 146097) / 4);
    my($d) = int((4 * ($c + 365)) / 1461) - 1;
    my($e) = $c - int((1461 * $d) / 4);
    my($m) = int((5 * ($e - 1) + 2) / 153);
    my($day) = $e - int((153 * $m + 2) / 5);
    my($month) = $m + 3 - 12 * int($m / 10);
    my($year) = $b * 100 + $d - 4800 + int($m / 10);

    return($year,$month,$day);

} #eosub--jd2ymd
#------------------------------------------------------------------------------
sub mjd2xte {
    # RXTE Mission Day Number: The number of 86400-second periods since
    # 1994-01-01 00:00:00. Note: RXTE Mission Days do not have leap seconds.

    my($mjd) = @_;

    return(&ckfloat($mjd) ? &shf2xte(&mjd2shf($mjd)) : 'error');

} #eosub--mjd2xte
#------------------------------------------------------------------------------
sub xte2mjd {

    my($xte) = @_;

    return(&ckfloat($xte) ? &shf2mjd(&xte2shf($xte)) : 'error');

} #eosub--xte2mjd
#------------------------------------------------------------------------------
sub xte2shf {

    my($xte) = @_;

    my($shfxte) = 441849609.0; # 1994-01-01 00:00:00 UTC(?) in SHF
    my($seconds_in_a_day) = 86400.0;

    return(&ckfloat($xte) ? $xte * $seconds_in_a_day + $shfxte : 'error');

} #eosub--xte2shf
#------------------------------------------------------------------------------
sub shf2xte {

    my($shf) = @_;

    my($shfxte) = 441849609.0; # 1994-01-01 00:00:00 UTC(?) in SHF
    my($seconds_in_a_day) = 86400.0;

    return(&ckfloat($shf) ? ($shf - $shfxte) / $seconds_in_a_day : 'error');

} #eosub--shf2xte
#------------------------------------------------------------------------------
sub mjd2shf {

    my($mjd) = @_;

    return('error') unless (&ckfloat($mjd));

    my($mjd0) = 44239.0; # 1980-01-01 00:00:00 UTC in MJD (UTC)
    my($seconds_in_a_day) = 86400.0;

    return((&leapsecs($mjd) - &leapsecs($mjd0))
	   + int(($mjd - $mjd0) * $seconds_in_a_day +
		 (($mjd >= $mjd0) ? 0.5 : -0.5)));

} #eosub--mjd2shf
#------------------------------------------------------------------------------
sub shf2mjd {

    my($shf) = @_;

    return('error') unless (&ckfloat($shf));

    my($mjd0) = 44239.0; # 1980-01-01 00:00:00 UTC in MJD (UTC)
    my($seconds_in_a_day) = 86400.0;

    return($mjd0 + (($shf - &leapsecs_shf($shf)) / $seconds_in_a_day));

} #eosub--shf2mjd
#------------------------------------------------------------------------------
sub mjd2time {

    # Converts MJD to Unix time_t format.
    my($mjd) = @_;

    return('error') unless (&ckfloat($mjd));

    # Due to POSIX bogosity, Unix does not track leap seconds.
    my($mjd_unix_epoch_start) = 40587.0; # 1970-01-01 00:00:00 UTC in MJD (UTC)
    my($seconds_in_a_day) = 86400.0;

    return(&nint(($mjd - $mjd_unix_epoch_start) * $seconds_in_a_day));

} #eosub--mjd2time
#------------------------------------------------------------------------------
sub time2mjd {

    # Converts Unix time_t format to MJD.
    my($time) = @_;

    return('error') unless (&ckfloat($time));

    # Due to POSIX bogosity, Unix does not track leap seconds.
    my($mjd_unix_epoch_start) = 40587.0; # 1970-01-01 00:00:00 UTC in MJD (UTC)
    my($seconds_in_a_day) = 86400.0;

    return($mjd_unix_epoch_start + ($time / $seconds_in_a_day));

} #eosub--time2mjd
#------------------------------------------------------------------------------
sub tt2utc_secs {

    #  Input: The number of seconds since any base epoch of your choosing in
    #         the TT system of time. You can optionally specify the base epoch
    #         MJD. If not specified, the current time is used.
    # Output: The number of seconds since any base epoch of your choosing in
    #         the UTC system of time.
    #   Note: This is valid only for TT times since the inception of UTC.
    my($tt_secs,$epoch) = @_;

    $epoch = &DateTime::now2mjd() unless ($epoch);
    return('error') if ($epoch < 41317.0);	# 1972-01-01
    return($tt_secs - 42.184 - &DateTime::leapsecs($epoch));

} #eosub--tt2utc_secs
#------------------------------------------------------------------------------
sub utc2tt_secs {

    #  Input: The number of seconds since any base epoch of your choosing in
    #         the UTC system of time. You can optionally specify the base epoch
    #         MJD. If not specified, the current time is used.
    # Output: The number of seconds since any base epoch of your choosing in
    #         the TT (Terrestrial Time) system of time.
    #   Note: This is valid only for UTC times since the inception of UTC.
    my($utc_secs,$epoch) = @_;

    $epoch = &DateTime::now2mjd() unless ($epoch);
    return('error') if ($epoch < 41317.0);	# 1972-01-01
    return($utc_secs + 42.184 + &DateTime::leapsecs($epoch));

} #eosub--utc2tt_secs
#------------------------------------------------------------------------------
sub tai2utc_secs {

    #  Input: The number of seconds since any base epoch of your choosing in
    #         the TAI (International Atomic Time) system of time. You can
    #         optionally specify the base epoch MJD. If not specified, the
    #         current time is used.
    # Output: The number of seconds since any base epoch of your choosing in
    #         the UTC system of time.
    #   Note: This is valid only for TAI times since the inception of UTC.
    my($tai_secs,$epoch) = @_;

    $epoch = &DateTime::now2mjd() unless ($epoch);
    return('error') if ($epoch < 41317.0);	# 1972-01-01
    return($tai_secs - 10 - &DateTime::leapsecs($epoch));

} #eosub--tai2utc_secs
#------------------------------------------------------------------------------
sub tai2tt_secs {

    #  Input: The number of seconds since any base epoch of your choosing in
    #         the TAI (International Atomic Time) system of time.
    # Output: The number of seconds since any base epoch of your choosing in
    #         the TT (Terrestrial Time) system of time.
    #   Note: This is probably valid only for TAI times since the inception
    #         of TAI in 1958. The actual difference varies slowly.
    my($tai_secs) = @_;

    return($tai_secs - 32.184);

} #eosub--tai2utc_secs
#------------------------------------------------------------------------------
sub leapsecs {

    my($mjd) = @_;

    # This function returns the number of leap seconds since 0 MJD.
    #
    # Assumptions: There is only one leap second and it is always positive.
    # Someday, one or both of these assumptions may become invalid, but for
    # now, all leap second adjustments so far have been +1 seconds.
    #
    # These are the exact MJD values of the various leap seconds. Actually,
    # the leap second is one second previous to these MJD values.
    # Source: United States Naval Observatory's Directorate of Time
    # See <URL:http://tycho.usno.navy.mil/leapsec.html> and
    # <URL:ftp://maia.usno.navy.mil/ser7/tai-utc.dat>.

    my(@mjd_leaps) = (
	41499.0, 41683.0, 42048.0, 42413.0, 42778.0, 43144.0, 43509.0,
	43874.0, 44239.0, 44786.0, 45151.0, 45516.0, 46247.0, 47161.0,
	47892.0, 48257.0, 48804.0, 49169.0, 49534.0, 50083.0, 50630.0,
	51179.0 );
    my($mjd_leap);
    my($num_secs) = 0;

    foreach $mjd_leap (@mjd_leaps) {
	if ($mjd >= $mjd_leap) {
	    $num_secs++;
	}
    }

    return($num_secs);

} #eosub--leapsecs
#------------------------------------------------------------------------------
sub leapsecs_shf {

    my($shf) = @_;

    # This function returns the number of leap seconds since SHF = 0;
    # if shf input value is negative, the number of leap seconds returned will
    # also be negative.
    
    my(@shfneg_leaps) = ( -236736009, -220838408, -189302407, -157766406,
			 -126230405, -94608004, -63072003, -31536002, -1 );
    my(@shfpos_leaps) = ( 47260800, 78796801, 110332802, 173491203, 252460804,
			 315619205, 347155206, 394416007, 425952008, 457488009,
			 504921610, 552182411, 599616012 );
    my($shf_leap);
    my($num_secs) = 0;

    if ($shf > 0) {
	foreach $shf_leap (@shfpos_leaps) {
	    if ($shf >= $shf_leap) {
		$num_secs++;
	    }
	}
    } elsif ($shf < 0) {
	foreach $shf_leap (@shfneg_leaps) {
	    if ($shf <= $shf_leap) {
		$num_secs--;
	    }
	}
    }

    return($num_secs);

} #eosub--leapsecs_shf
#------------------------------------------------------------------------------
sub nint {

    my($value) = @_;

    return(int(($value >= 0.0) ? $value + 0.5 : $value - 0.5));

} #eosub--nint
#------------------------------------------------------------------------------
sub ckint {

    return($_[0] =~ /^\s*[+-]?\d+\s*$/);

} #eosub--ckint
#------------------------------------------------------------------------------
sub ckfloat {

    return($_[0] =~ /^\s*[+-]?\d+(\.\d+)?([eE][+-]?\d{1,3})?\s*$/);

} #eosub--ckfloat
#------------------------------------------------------------------------------
sub ckymd {

    my($y,$m,$d) = @_;

    return(0) unless (&ckint($y) && &ckint($m) && &ckint($d));

    my(@dom) = (0,31,28,31,30,31,30,31,31,30,31,30,31);
    $dom[2] += &leapyear($y);

    return(($m > 0) && ($m <= 12) && ($d > 0) && ($d <= $dom[$m]));

} #eosub--ckymd
#------------------------------------------------------------------------------
1;
